from typing import Tuple

import numpy as np
import xarray as xr

from starfish.core.codebook.codebook import Codebook
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.imagestack.test.factories import create_imagestack_from_codebook
from starfish.core.spots.DecodeSpots import MetricDistance, PerRoundMaxChannel, SimpleLookupDecoder
from starfish.core.spots.FindSpots import BlobDetector
from starfish.core.types import Axes, Features


def simple_gaussian_spot_detector() -> BlobDetector:
    """create a basic gaussian spot detector"""
    return BlobDetector(
        min_sigma=1,
        max_sigma=4,
        num_sigma=5,
        threshold=0,
        measurement_type='max')


def composite_codebook() -> Tuple[Codebook, ImageStack]:
    """
    Produce an Imagestack representing a composite experiment where the first 2 rounds are
    multiplexed data and the last round is sequential data.

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    """
    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 2, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_C"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 2, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_D"
        }
    ]

    codebook = Codebook.from_code_array(codebook_data)
    imagestack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10), (6, 90, 10), (7, 90, 10)),
        codebook=codebook
    )
    return codebook, imagestack


def composite_codebook_mixed_round() -> Tuple[Codebook, ImageStack]:
    """
    Produce an Imagestack representing a composite experiment where the first 2 and a half rounds
    are multiplexed data and the last round and a half is sequential data. This represents the type
    of hybrid experiment happening at the allen.

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    """
    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 2, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_C"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 2, Axes.CH.value: 2, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_D"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 3, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_E"
        },
    ]

    codebook = Codebook.from_code_array(codebook_data)
    imagestack = create_imagestack_from_codebook(
        pixel_dimensions=(1, 10, 10),
        spot_coordinates=((0, 5, 6), (0, 2, 3), (0, 7, 1), (0, 8, 2), (0, 3, 6)),
        codebook=codebook
    )
    return codebook, imagestack


def compostie_codebook_seperate_stacks() -> Tuple[Codebook, ImageStack, ImageStack]:
    """
    Produce separate different sized ImageStacks containing data from two different experiment
    types (multiplexed and non multiplexed) and one codebook with the information for both.

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    """

    # 3 round 3 ch multiplexed data
    multiplexed_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 2, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(multiplexed_data)
    multiplexed_stack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    # 3 rounds 1 ch non multiplexed data
    sequential_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 3, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_C"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 4, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_D"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 5, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_D"
        }
    ]
    codebook = Codebook.from_code_array(sequential_data)
    sequential_stack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10), (7, 90, 10)),
        codebook=codebook
    )
    sequential_stack = sequential_stack.sel(indexers={Axes.ROUND: (3, 6)})

    # create codebook with combined target values
    combined = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 2, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 3, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_C"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 4, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_D"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 5, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_E"
        }

    ]
    codebook = Codebook.from_code_array(combined)
    return codebook, multiplexed_stack, sequential_stack


gaussian_spot_detector = simple_gaussian_spot_detector()


def test_composite_codebook_decoding():
    """
    Test a decoding workflow in which one ImageStack is composed of both multiplexed data and
    sequential data in separate rounds. Show that it's easier to select our the two different
    sections and process them separately.
    """
    # create an ImageStack where rounds 0,1 contain multiplexed data and round 2 contains
    # sequential data
    codebook, stack = composite_codebook()

    # select out the different portions of the imagestack
    multiplexed_stack = stack.sel(indexers={Axes.ROUND: (0, 1)})
    sequential_stack = stack.sel(indexers={Axes.ROUND: 2})

    # select out the corresponding portions of the codebooks
    codebook_multiplexed = codebook.get_partial(indexers={Axes.ROUND: (0, 1)})
    codebook_sequential = codebook.get_partial(indexers={Axes.ROUND: 2})

    # find spots
    reference_image = multiplexed_stack.reduce((Axes.ROUND, Axes.CH), func="max")
    multiplexed_spots = gaussian_spot_detector.run(
        image_stack=multiplexed_stack, reference_image=reference_image)

    sequential_spots = gaussian_spot_detector.run(image_stack=sequential_stack)

    # decode
    per_round_decoder = PerRoundMaxChannel(codebook=codebook_multiplexed)
    decoded_multiplexed = per_round_decoder.run(spots=multiplexed_spots)
    assert np.array_equal(decoded_multiplexed[Features.TARGET].values, ['GENE_B', 'GENE_A'])

    simple_lookup_decoder = SimpleLookupDecoder(codebook=codebook_sequential)
    decoded_sequential = simple_lookup_decoder.run(spots=sequential_spots)
    assert np.array_equal(decoded_sequential[Features.TARGET].values, ['GENE_C', 'GENE_D'])


def test_composite_codebook_decoding_seperate_sized_stacks():
    """
    Test a decoding workflow in which two different ImageStacks of differing sizes and experiment
    types are created from the same codebook. Show that it's possible to select out the portion of
    the codebook to use successfully for each ImageStack.
    """

    # create two different sized ImageStacks with the same codebook
    codebook, multiplexed_stack, sequential_stack = compostie_codebook_seperate_stacks()
    codebook_multiplexed = codebook.get_partial(indexers={Axes.ROUND: (0, 2)})

    # find spots
    reference_image = multiplexed_stack.reduce((Axes.ROUND, Axes.CH), func="max")
    multiplexed_spots = gaussian_spot_detector.run(
        image_stack=multiplexed_stack, reference_image=reference_image)

    sequential_spots = gaussian_spot_detector.run(image_stack=sequential_stack)

    metric_decoder = MetricDistance(codebook=codebook_multiplexed,
                                    max_distance=0,
                                    min_intensity=.1,
                                    norm_order=2)

    decoded_multiplexed = metric_decoder.run(spots=multiplexed_spots)

    assert np.array_equal(decoded_multiplexed[Features.TARGET].values, ['GENE_B', 'GENE_A'])

    lookup_decoder = SimpleLookupDecoder(codebook=codebook)
    decoded_sequential = lookup_decoder.run(spots=sequential_spots)

    assert np.array_equal(sorted(decoded_sequential[Features.TARGET].values),
                          ['GENE_C', 'GENE_D', 'GENE_E'])


def test_composite_codebook_mixed_round():
    """
    Test a decoding workflow in which one ImageStack is composed of both multiplexed data and
    sequential data with one round having both. Show that a decoding workflow is possible by
    selecting out three different sections of the stack and processing them separately.
    """
    codebook, stack = composite_codebook_mixed_round()

    # select the multiplexed portion of the stack
    multiplexed_stack = stack.sel(indexers={Axes.ROUND: (0, 2), Axes.CH: (0, 1)})

    # select the the sequential portion of the combined round
    sequential_stack_part_1 = stack.sel(indexers={Axes.ROUND: 2, Axes.CH: (1, 2)})
    # select the rest of the sequential portion of the stack
    sequential_stack_part_2 = stack.sel(indexers={Axes.ROUND: 3, Axes.CH: (0, 2)})

    # select out the multiplexed portion of the codebook
    codebook_multiplexed = codebook.get_partial(indexers={Axes.ROUND: (0, 2), Axes.CH: (0, 1)})

    # find spots
    reference_image = multiplexed_stack.reduce((Axes.ROUND, Axes.CH), func="max")
    multiplexed_spots = gaussian_spot_detector.run(
        image_stack=multiplexed_stack, reference_image=reference_image)

    # the easiest way to process the sequential portion is to process the two parts independently
    sequential_spots_part_1 = gaussian_spot_detector.run(image_stack=sequential_stack_part_1)
    sequential_spots_part_2 = gaussian_spot_detector.run(image_stack=sequential_stack_part_2)

    metric_decoder = MetricDistance(codebook=codebook_multiplexed,
                                    max_distance=0,
                                    min_intensity=.1,
                                    norm_order=2)

    decoded_multiplexed = metric_decoder.run(spots=multiplexed_spots)

    assert np.array_equal(sorted(decoded_multiplexed[Features.TARGET].values),
                          ['GENE_A', 'GENE_B', 'GENE_C'])

    # decode sequential portions independently then combine the results
    lookup_decoder = SimpleLookupDecoder(codebook=codebook)
    decoded_sequential_part_1 = lookup_decoder.run(spots=sequential_spots_part_1)
    decoded_sequential_part_2 = lookup_decoder.run(spots=sequential_spots_part_2)
    decoded_sequential = xr.concat([decoded_sequential_part_1, decoded_sequential_part_2],
                                   dim="features")

    assert np.array_equal(sorted(decoded_sequential[Features.TARGET].values), ['GENE_D', 'GENE_E'])



from typing import Iterable

import xarray as xr

from starfish.core.types import Features
from .expression_matrix import ExpressionMatrix


def concatenate(expression_matrices: Iterable[ExpressionMatrix]) -> ExpressionMatrix:
    """Concatenate IntensityTables produced for different fields of view or across imaging rounds

    Expression Matrices are concatenated along the cells axis, and the resulting arrays are stored
    densely.

    Parameters
    ----------
    expression_matrices : Iterable[ExpressionMatrix]
        iterable (list-like) of expression matrices to combine

    Returns
    -------
    ExpressionMatrix :
        Concatenated expression matrix containing all input cells. Missing gene values are filled
        with np.nan

    See Also
    --------
    Combine_first: http://xarray.pydata.org/en/stable/combining.html#combine

    """
    concatenated_matrix: xr.DataArray = xr.concat(list(expression_matrices), Features.CELLS)
    return ExpressionMatrix(concatenated_matrix)



import os
from pathlib import Path
from typing import Mapping, Sequence

import numpy as np
from slicedimage import ImageFormat

from starfish.core.experiment.experiment import Experiment, FieldOfView
from starfish.core.imagestack.test.imagestack_test_utils import verify_physical_coordinates
from starfish.core.types import Axes, Coordinates, CoordinateValue
from .util import (
    render_coordinates_to_rows,
    UniqueTiles,
    write_coordinates_csv,
    write_tile_data,
)
from ..factories import unique_data
from ... import (
    tile_fetcher_factory,
    TileFetcher,
    TileIdentifier,
)
from ...structured_formatter import format_structured_dataset


def test_multiple_aligned_regular_fov(
        tmpdir,
        fovs=(0, 1, 2, 3),
        rounds=(1, 2, 4),
        chs=(2, 3, 4),
        zplanes=(0, 1, 2),
        tile_height=100,
        tile_width=60,
        fov_to_x_coords={
            0: (0.0, 0.1),
            1: (0.0, 0.1),
            2: (0.1, 0.2),
            3: (0.1, 0.2),
        },
        fov_to_y_coords={
            0: (0.0, 0.1),
            1: (0.1, 0.2),
            2: (0.0, 0.1),
            3: (0.1, 0.2),
        },
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default value is mutable, but it's for readability reasons!
):
    """Write the tiles for a multi-fov aligned (physical coordinates) regular (the dimensions have
    the same cardinality) image.  Then build an experiment from the tiles.  Finally, load the
    resulting experiment as an ImageStack and verify that the data matches."""
    tmpdir_path: Path = Path(tmpdir)
    tile_identifiers: Sequence[TileIdentifier] = [
        TileIdentifier(fov_id, round_label, ch_label, zplane_label)
        for fov_id in fovs
        for round_label in rounds
        for ch_label in chs
        for zplane_label in zplanes
    ]
    tile_fetcher: TileFetcher = tile_fetcher_factory(
        UniqueTiles,
        pass_tile_indices=True,
        fovs=fovs,
        rounds=rounds,
        chs=chs,
        zplanes=zplanes,
        tile_height=tile_height,
        tile_width=tile_width,
    )
    tile_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = {
        tile_identifier: {
            Coordinates.X: fov_to_x_coords[tile_identifier.fov_id],
            Coordinates.Y: fov_to_y_coords[tile_identifier.fov_id],
            Coordinates.Z: zplane_to_coords[tile_identifier.zplane_label],
        }
        for tile_identifier in tile_identifiers
    }

    write_tile_data(
        tmpdir_path,
        FieldOfView.PRIMARY_IMAGES,
        ImageFormat.TIFF,
        tile_identifiers,
        tile_fetcher)

    coordinates_csv_path = tmpdir_path / "coordinates.csv"
    rows = render_coordinates_to_rows(tile_coordinates)
    write_coordinates_csv(coordinates_csv_path, rows, True)

    format_structured_dataset(
        os.fspath(tmpdir_path),
        os.fspath(coordinates_csv_path),
        os.fspath(tmpdir_path),
        ImageFormat.TIFF,
        False,
    )

    # load the data and verify it.
    exp = Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))

    for fov_id in fovs:
        fov = exp.fov(lambda fieldofview: fieldofview.name == f"fov_{fov_id:03}")
        stack = fov.get_image(FieldOfView.PRIMARY_IMAGES)
        for round_label in rounds:
            for ch_label in chs:
                for zplane_label in zplanes:
                    data, _ = stack.get_slice({
                        Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label
                    })
                    expected_data = unique_data(
                        fovs.index(fov_id),
                        rounds.index(round_label),
                        chs.index(ch_label),
                        zplanes.index(zplane_label),
                        len(fovs), len(rounds), len(chs), len(zplanes),
                        tile_height, tile_width,
                    )
                    assert np.allclose(data, expected_data)

        for selectors in stack._iter_axes({Axes.ZPLANE}):
            zplane_label = selectors[Axes.ZPLANE]
            verify_physical_coordinates(
                stack,
                fov_to_x_coords[fov_id],
                fov_to_y_coords[fov_id],
                zplane_to_coords[zplane_label],
                selectors[Axes.ZPLANE])



from pathlib import Path
from typing import Any, Hashable, Mapping, MutableMapping, Optional, Tuple, Union

import numpy as np
import xarray as xr
from semantic_version import Version

from starfish.core.morphology.util import (
    _get_axes_names,
    _normalize_physical_ticks,
    _normalize_pixel_ticks,
)
from starfish.core.types import ArrayLike, Axes, Coordinates, LOG, Number, STARFISH_EXTRAS_KEY
from starfish.core.util.logging import Log


class AttrKeys:
    DOCTYPE = f"{STARFISH_EXTRAS_KEY}.DOCTYPE"
    LOG = f"{STARFISH_EXTRAS_KEY}.{LOG}"
    VERSION = f"{STARFISH_EXTRAS_KEY}.VERSION"


DOCTYPE_STRING = "starfish/LabelImage"
CURRENT_VERSION = Version("0.0.0")
MIN_SUPPORTED_VERSION = Version("0.0.0")
MAX_SUPPORTED_VERSION = Version("0.0.0")


class LabelImage:
    """Wraps an xarray that contains a 2D or 3D labeled image.  Each axis is labeled with physical
    coordinate data."""

    def __init__(self, label_image: xr.DataArray):
        # verify that the data array has the required elements.
        if label_image.dtype.kind not in ("i", "u"):
            raise TypeError("label image should be an integer type")
        for axis in (Axes.X, Axes.Y):
            if axis.value not in label_image.coords:
                raise ValueError(f"label image should have an {axis.value} axis")
        expected_physical_coordinates: Tuple[Coordinates, ...]
        if label_image.ndim == 5:
            expected_physical_coordinates = (Coordinates.X, Coordinates.Y, Coordinates.Z)
        else:
            expected_physical_coordinates = (Coordinates.X, Coordinates.Y)
        for coord in expected_physical_coordinates:
            if coord.value not in label_image.coords:
                raise ValueError(f"label image should have a {coord.value} coordinates")

        self.label_image = label_image.copy(deep=False)
        if AttrKeys.DOCTYPE not in self.label_image.attrs:
            self.label_image.attrs[AttrKeys.DOCTYPE] = DOCTYPE_STRING
        if AttrKeys.LOG not in self.label_image.attrs:
            self.label_image.attrs[AttrKeys.LOG] = Log().encode()

    @classmethod
    def from_label_array_and_ticks(
            cls,
            array: np.ndarray,
            pixel_ticks: Optional[Union[
                Mapping[Axes, ArrayLike[int]],
                Mapping[str, ArrayLike[int]]]],
            physical_ticks: Union[
                Mapping[Coordinates, ArrayLike[Number]],
                Mapping[str, ArrayLike[Number]]],
            log: Optional[Log],
    ) -> "LabelImage":
        """Constructs a LabelImage from an array containing the labels, a set of physical
        coordinates, and an optional log of how this label image came to be.

        Parameters
        ----------
        array : np.ndarray
            A 2D or 3D array containing the labels.  The ordering of the axes must be Y, X for 2D
            images and ZPLANE, Y, X for 3D images.
        pixel_ticks : Optional[Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]]]
            A map from the axis to the values for that axis.  For any axis that exist in the array
            but not in pixel_coordinates, the pixel coordinates are assigned from 0..N-1, where N is
            the size along that axis.
        physical_ticks : Union[Mapping[Coordinates, ArrayLike[Number]], Mapping[str,
        ArrayLike[Number]]]
            A map from the physical coordinate type to the values for axis.  For 2D label images,
            X and Y physical coordinates must be provided.  For 3D label images, Z physical
            coordinates must also be provided.
        log : Optional[Log]
            A log of how this label image came to be.
        """
        # normalize the pixel coordinates to Mapping[Axes, ArrayLike[int]]
        pixel_ticks = _normalize_pixel_ticks(pixel_ticks)
        # normalize the physical coordinates to Mapping[Coordinates, ArrayLike[Number]]
        physical_ticks = _normalize_physical_ticks(physical_ticks)

        img_axes, img_coords = _get_axes_names(array.ndim)
        xr_axes = [axis.value for axis in img_axes]
        try:
            xr_coords: MutableMapping[Hashable, Any] = {
                coord.value: (axis.value, physical_ticks[coord])
                for axis, coord in zip(img_axes, img_coords)
            }
        except KeyError as ex:
            raise KeyError(f"missing physical coordinates {ex.args[0]}") from ex

        for ix, axis in enumerate(img_axes):
            xr_coords[axis.value] = pixel_ticks.get(axis, np.arange(0, array.shape[ix]))

        dataarray = xr.DataArray(
            array,
            dims=xr_axes,
            coords=xr_coords,
        )
        dataarray.attrs.update({
            AttrKeys.LOG: (log or Log()).encode(),
            AttrKeys.DOCTYPE: DOCTYPE_STRING,
            AttrKeys.VERSION: str(CURRENT_VERSION),
        })

        return LabelImage(dataarray)

    @property
    def xarray(self):
        """Returns the xarray that contains the label image and the physical coordinates."""
        return self.label_image

    @property
    def log(self) -> Log:
        """Returns a copy of the provenance data.  Modifications to this copy will not affect the
        log stored on this label image."""
        return Log.decode(self.label_image.attrs[AttrKeys.LOG])

    @classmethod
    def open_netcdf(cls, path: Union[str, Path]) -> "LabelImage":
        """Load a label image saved as a netcdf file from disk.

        Parameters
        ----------
        path : Union[str, Path]
            Path of the label image to instantiate from.

        Returns
        -------
        label_image : LabelImage
            Label image from the path.
        """
        label_image = xr.open_dataarray(path)
        if (
                AttrKeys.DOCTYPE not in label_image.attrs
                or label_image.attrs[AttrKeys.DOCTYPE] != DOCTYPE_STRING
                or AttrKeys.VERSION not in label_image.attrs
        ):
            raise ValueError(f"{path} does not appear to be a starfish label image")
        if not (
                MIN_SUPPORTED_VERSION
                <= Version(label_image.attrs[AttrKeys.VERSION])
                <= MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"{path} contains a label image, but the version "
                f"{label_image.attrs[AttrKeys.VERSION]} is not supported")

        return cls(label_image)

    def to_netcdf(self, path: Union[str, Path]):
        """Save the label image as a netcdf file.

        Parameters
        ----------
        path : Union[str, Path]
            Path of the netcdf file to write to.
        """
        self.label_image.to_netcdf(path)



import os

import pandas as pd

import starfish
from starfish.core.test.factories import two_spot_sparse_coded_data_factory


def test_per_round_max_spot_decoding_without_spots(tmpdir):

    codebook, image_stack, max_intensity = two_spot_sparse_coded_data_factory()

    bd = starfish.spots.FindSpots.BlobDetector(
        min_sigma=1, max_sigma=1, num_sigma=1, threshold=max_intensity + 0.1)
    no_spots = bd.run(image_stack)

    decode = starfish.spots.DecodeSpots.PerRoundMaxChannel(codebook)
    decoded_no_spots: starfish.DecodedIntensityTable = decode.run(spots=no_spots)

    decoded_spot_table = decoded_no_spots.to_decoded_dataframe()

    filename = os.path.join(tmpdir, 'test.csv')
    decoded_spot_table.save_csv(os.path.join(tmpdir, 'test.csv'))

    # verify we can concatenate two empty tables
    table1 = pd.read_csv(filename, index_col=0)
    table2 = pd.read_csv(filename, index_col=0)
    pd.concat([table1, table2], axis=0)


def test_metric_decoding_without_spots(tmpdir):

    codebook, image_stack, max_intensity = two_spot_sparse_coded_data_factory()

    bd = starfish.spots.FindSpots.BlobDetector(
        min_sigma=1, max_sigma=1, num_sigma=1, threshold=max_intensity + 0.1)
    no_spots = bd.run(image_stack)

    decode = starfish.spots.DecodeSpots.MetricDistance(
        codebook, max_distance=0, min_intensity=max_intensity + 0.1
    )
    decoded_no_spots: starfish.DecodedIntensityTable = decode.run(spots=no_spots)

    decoded_spot_table = decoded_no_spots.to_decoded_dataframe()

    filename = os.path.join(tmpdir, 'test.csv')
    decoded_spot_table.save_csv(os.path.join(tmpdir, 'test.csv'))

    # verify we can concatenate two empty tables
    table1 = pd.read_csv(filename, index_col=0)
    table2 = pd.read_csv(filename, index_col=0)
    pd.concat([table1, table2], axis=0)



from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr
from scipy.signal import convolve, fftconvolve

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number
from ._base import FilterAlgorithm
from .util import (
    determine_axes_to_group_by,
    gaussian_kernel,
)


class DeconvolvePSF(FilterAlgorithm):
    """
    Deconvolve a point spread function from the image. The point spread function is assumed to be
    an isotropic Gaussian, with a user specified standard deviation, sigma.

    There are currently several issues with this function.
    See https://github.com/spacetx/starfish/issues/731

    Parameters
    ----------
    num_iter : int
        number of iterations to run. Note that this is a very important parameter that requires
        careful optimization
    sigma : Number
        standard deviation of the gaussian kernel used to construct the point spread function
    is_volume: bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.

    Examples
    --------
    >>> from skimage import color, data, restoration
    >>> camera = color.rgb2gray(data.camera())
    >>> from scipy.signal import convolve2d
    >>> psf = np.ones((5, 5)) / 25
    >>> camera = convolve2d(camera, psf, 'same')
    >>> camera += 0.1 * camera.std() * np.random.standard_normal(camera.shape)
    >>> deconvolved = restoration.richardson_lucy(camera, psf, 5)

    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Richardson%E2%80%93Lucy_deconvolution

    Notes
    -----
    This code is based on code from skimage.restoration. We modified it to implement a bugfix
    wherein zeros in the input image or zeros produced during an intermediate would induce
    divide by zero -> Nan. These Nans would then propagate throughout the image, invalidating
    the results. Longer term, we will make a PR to skimage to introduce the fix. There is some
    existing work linked here: https://github.com/scikit-image/scikit-image/issues/2551

    """

    def __init__(
            self,
            num_iter: int,
            sigma: Number,
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:

        self.num_iter = num_iter
        self.sigma = sigma
        self.kernel_size: int = int(2 * np.ceil(2 * sigma) + 1)
        self.psf: np.ndarray = gaussian_kernel(
            shape=(self.kernel_size, self.kernel_size),
            sigma=sigma
        )
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"num_iter": 2, "sigma": 1}

    # Here be dragons. This algorithm had a bug, but the results looked nice. Now we've "fixed" it
    # and the results look bad. #548 addresses this problem.
    @staticmethod
    def _richardson_lucy_deconv(
            image: xr.DataArray, iterations: int, psf: np.ndarray
    ) -> xr.DataArray:
        """
        Deconvolves input image with a specified point spread function.

        Parameters
        ----------
        image : xr.DataArray
           Input degraded image (can be N dimensional).
        psf : ndarray
           The point spread function.
        iterations : int
           Number of iterations. This parameter plays the role of
           regularisation.

        Returns
        -------
        im_deconv : xr.DataArray
           The deconvolved image.

        """
        # compute the times for direct convolution and the fft method. The fft is of
        # complexity O(N log(N)) for each dimension and the direct method does
        # straight arithmetic (and is O(n*k) to add n elements k times)
        direct_time = np.prod(image.shape + psf.shape)
        fft_time = np.sum([n * np.log(n) for n in image.shape + psf.shape])

        # see whether the fourier transform convolution method or the direct
        # convolution method is faster (discussed in scikit-image PR #1792)
        time_ratio = 40.032 * fft_time / direct_time

        if time_ratio <= 1 or len(image.shape) > 2:
            convolve_method = fftconvolve
        else:
            convolve_method = convolve

        image = image.astype(np.float)
        psf = psf.astype(np.float)
        im_deconv = 0.5 * np.ones(image.shape)
        psf_mirror = psf[::-1, ::-1]

        eps = np.finfo(image.dtype).eps
        for _ in range(iterations):
            x = convolve_method(im_deconv, psf, 'same')
            np.place(x, x == 0, eps)
            relative_blur = image / x + eps
            im_deconv *= convolve_method(relative_blur, psf_mirror, 'same')

        if np.all(np.isnan(im_deconv)):
            raise RuntimeError(
                'All-NaN output data detected. Likely cause is that deconvolution has been run for '
                'too many iterations.')

        return im_deconv

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose=False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        func = partial(
            self._richardson_lucy_deconv,
            iterations=self.num_iter, psf=self.psf
        )
        result = stack.apply(
            func,
            group_by=group_by,
            verbose=verbose,
            n_processes=n_processes,
            in_place=in_place,
            level_method=self.level_method,
        )
        return result



from typing import Mapping, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.morphology.util import _get_axes_names
from starfish.core.types import ArrayLike, Axes, Coordinates, Number
from ._base import BinarizeAlgorithm


class ThresholdBinarize(BinarizeAlgorithm):
    """Binarizes an image using a threshold.  Pixels that exceed the threshold are considered True
    and all remaining pixels are considered False.

    The image being binarized must be an ImageStack with num_rounds == 1 and num_chs == 1.
    """
    def __init__(self, threshold: Number):
        self.threshold = threshold

    def _binarize(self, result: np.ndarray, tile_data: Union[np.ndarray, xr.DataArray]) -> None:
        result[:] = np.asarray(tile_data) >= self.threshold

    def run(self, image: ImageStack, *args, **kwargs) -> BinaryMaskCollection:
        if image.num_rounds != 1:
            raise ValueError(
                f"{ThresholdBinarize.__name__} given an image with more than one round "
                f"{image.num_rounds}")
        if image.num_chs != 1:
            raise ValueError(
                f"{ThresholdBinarize.__name__} given an image with more than one channel "
                f"{image.num_chs}")

        result_array = np.empty(
            shape=[
                image.shape[axis]
                for axis, _ in zip(*_get_axes_names(3))
            ],
            dtype=np.bool)

        # TODO: (ttung) This could theoretically be done with ImageStack.transform, but
        # ImageStack.transform doesn't provide the selectors to the worker method.  In this case,
        # we need the selectors to select the correct region of the output array.  The alternative
        # is for each worker thread to create a new array, and then merge them at the end, but that
        # effectively doubles our memory consumption.
        #
        # For now, we will just do it in-process, because it's not a particularly compute-intensive
        # task.
        self._binarize(result_array, image.xarray[0, 0])

        pixel_ticks: Mapping[Axes, ArrayLike[int]] = {
            Axes(axis): axis_data
            for axis, axis_data in image.xarray.coords.items()
            if axis in _get_axes_names(3)[0]
        }
        physical_ticks: Mapping[Coordinates, ArrayLike[Number]] = {
            Coordinates(coord): coord_data
            for coord, coord_data in image.xarray.coords.items()
            if coord in _get_axes_names(3)[1]
        }

        return BinaryMaskCollection.from_binary_arrays_and_ticks(
            (result_array,),
            pixel_ticks,
            physical_ticks,
            image.log,
        )



"""These are tests that place the tiles in different locations that are still considered valid.
For instance, primary-f0_r000000_c0_z0.tiff should still resolve to tile in round 0, ch 0, zplane 0
in FOV 0.  We should also be able to put tiles in arbitrary subdirectories and still have everything
work.
"""

import os
import random
import string
from pathlib import Path
from typing import Callable, Mapping, Sequence, Tuple

import numpy as np
from slicedimage import ImageFormat

from starfish.core.experiment.experiment import Experiment, FieldOfView
from starfish.core.imagestack.test.imagestack_test_utils import verify_physical_coordinates
from starfish.core.types import Axes, Coordinates, CoordinateValue, Number
from .util import (
    render_coordinates_to_rows,
    UniqueTiles,
    write_coordinates_csv,
    write_tile_data,
)
from ..factories import unique_data
from ... import (
    tile_fetcher_factory,
    TileFetcher,
    TileIdentifier,
)
from ...structured_formatter import format_structured_dataset


def tilepath_test(
        tmpdir_path: Path,
        tilepath_generator: Callable[[Path, str, ImageFormat, TileIdentifier], Path],
        rounds: Sequence[int],
        chs: Sequence[int],
        zplanes: Sequence[int],
        tile_height: int,
        tile_width: int,
        x_coords: Tuple[int, int],
        y_coords: Tuple[int, int],
        zplane_to_coords: Mapping[int, Number],
):
    """Write the tiles for a single aligned (physical coordinates) regular (the dimensions have the
    same cardinality) FOV, with a provided tile_path_generator.  Then build an experiment from the
    tiles.  Finally, load the resulting experiment as an ImageStack and verify that the data
    matches."""
    tile_identifiers: Sequence[TileIdentifier] = [
        TileIdentifier(0, round_label, ch_label, zplane_label)
        for round_label in rounds
        for ch_label in chs
        for zplane_label in zplanes
    ]
    tile_fetcher: TileFetcher = tile_fetcher_factory(
        UniqueTiles,
        pass_tile_indices=True,
        fovs=[0],
        rounds=rounds,
        chs=chs,
        zplanes=zplanes,
        tile_height=tile_height,
        tile_width=tile_width,
    )
    tile_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = {
        tile_identifier: {
            Coordinates.X: x_coords,
            Coordinates.Y: y_coords,
            Coordinates.Z: zplane_to_coords[tile_identifier.zplane_label],
        }
        for tile_identifier in tile_identifiers
    }

    write_tile_data(
        tmpdir_path,
        FieldOfView.PRIMARY_IMAGES,
        ImageFormat.TIFF,
        tile_identifiers,
        tile_fetcher,
        tilepath_generator,
    )

    coordinates_csv_path = tmpdir_path / "coordinates.csv"
    rows = render_coordinates_to_rows(tile_coordinates)
    write_coordinates_csv(coordinates_csv_path, rows, True)

    format_structured_dataset(
        os.fspath(tmpdir_path),
        os.fspath(coordinates_csv_path),
        os.fspath(tmpdir_path),
        ImageFormat.TIFF,
        False,
    )

    # load the data and verify it.
    exp = Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))
    fov = exp.fov()
    stack = fov.get_image(FieldOfView.PRIMARY_IMAGES)

    for round_label in rounds:
        for ch_label in chs:
            for zplane_label in zplanes:
                data, _ = stack.get_slice({
                    Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label
                })
                expected_data = unique_data(
                    0, rounds.index(round_label), chs.index(ch_label), zplanes.index(zplane_label),
                    1, len(rounds), len(chs), len(zplanes),
                    tile_height, tile_width,
                )
                assert np.allclose(data, expected_data)

    for selectors in stack._iter_axes({Axes.ZPLANE}):
        zplane_label = selectors[Axes.ZPLANE]
        verify_physical_coordinates(
            stack, x_coords, y_coords, zplane_to_coords[zplane_label], selectors[Axes.ZPLANE])


def test_zeropadded_filenames(
        tmpdir,
        rounds=(1, 2, 4),
        chs=(2, 3, 4),
        zplanes=(0, 1, 2),
        tile_height=100,
        tile_width=60,
        x_coords=(0.0, 0.1),
        y_coords=(0.1, 0.2),
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default value is mutable, but it's for readability reasons!
):
    tmpdir_path = Path(tmpdir)
    def tilepath_generator(
            basepath: Path,
            image_type: str,
            tile_format: ImageFormat,
            tile_identifier: TileIdentifier,
    ) -> Path:
        return basepath / (
            f"{image_type}-"
            f"f{tile_identifier.fov_id}-"
            f"r{tile_identifier.round_label:02}-"
            f"c{tile_identifier.ch_label}-"
            f"z{tile_identifier.zplane_label}."
            f"{tile_format.file_ext}"
        )

    tilepath_test(
        tmpdir_path,
        tilepath_generator,
        rounds,
        chs,
        zplanes,
        tile_height,
        tile_width,
        x_coords,
        y_coords,
        zplane_to_coords,
    )


def test_tiles_in_dirs(
        tmpdir,
        rounds=(1, 2, 4),
        chs=(2, 3, 4),
        zplanes=(0, 1, 2),
        tile_height=100,
        tile_width=60,
        x_coords=(0.0, 0.1),
        y_coords=(0.1, 0.2),
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default value is mutable, but it's for readability reasons!
):
    tmpdir_path = Path(tmpdir)
    def tilepath_generator(
            basepath: Path,
            image_type: str,
            tile_format: ImageFormat,
            tile_identifier: TileIdentifier,
    ) -> Path:
        rand_subdir = "".join([
            random.choice(string.ascii_letters + string.digits)
            for n in range(16)])
        return basepath / rand_subdir / (
            f"{image_type}-"
            f"f{tile_identifier.fov_id}-"
            f"r{tile_identifier.round_label:02}-"
            f"c{tile_identifier.ch_label}-"
            f"z{tile_identifier.zplane_label}."
            f"{tile_format.file_ext}"
        )

    tilepath_test(
        tmpdir_path,
        tilepath_generator,
        rounds,
        chs,
        zplanes,
        tile_height,
        tile_width,
        x_coords,
        y_coords,
        zplane_to_coords,
    )



import numpy as np
import pytest

from starfish.core.types import Axes, Coordinates
from .factories import label_array_2d, label_array_3d
from ..binary_mask import BinaryMaskCollection


def test_2d():
    """Simple case of BinaryMaskCollection.from_label_array_and_ticks with 2D data.  Pixel ticks are
    inferred."""
    label_array, physical_ticks = label_array_2d()

    binary_mask_collection = BinaryMaskCollection.from_label_array_and_ticks(
        label_array,
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 2

    region_0, region_1 = binary_mask_collection.masks()
    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 6), dtype=np.bool))
    temp = np.ones((2, 3), dtype=np.bool)
    temp[-1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [3, 4, 5])

    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_3d():
    """Simple case of BinaryMaskCollection.from_label_array_and_ticks with 3D data.  Pixel ticks are
    inferred."""
    label_array, physical_ticks = label_array_3d()

    binary_mask_collection = BinaryMaskCollection.from_label_array_and_ticks(
        label_array,
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 2

    region_0, region_1 = binary_mask_collection.masks()
    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 1, 6), dtype=np.bool))
    temp = np.ones((2, 2, 3), dtype=np.bool)
    temp[-1, -1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.ZPLANE.value], [0])
    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.ZPLANE.value], [0, 1])
    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [3, 4, 5])

    assert np.array_equal(region_0[Coordinates.Z.value],
                          physical_ticks[Coordinates.Z][0:1])
    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Z.value],
                          physical_ticks[Coordinates.Z][0:2])
    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_from_label_array_provided_pixel_ticks():
    """BinaryMaskCollection.from_label_array_and_ticks with 2D data and some pixel ticks
    provided."""
    label_array, physical_ticks = label_array_2d()
    pixel_ticks = {
        Axes.X: [2, 3, 4, 5, 6, 7],
    }

    binary_mask_collection = BinaryMaskCollection.from_label_array_and_ticks(
        label_array,
        pixel_ticks,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 2

    region_0, region_1 = binary_mask_collection.masks()
    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 6), dtype=np.bool))
    temp = np.ones((2, 3), dtype=np.bool)
    temp[-1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [2, 3, 4, 5, 6, 7])

    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [5, 6, 7])

    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_incorrectly_sized_pixel_ticks():
    """BinaryMaskCollection.from_label_array_and_ticks with 2D data and some pixel ticks provided,
    albeit of the wrong cardinality."""
    label_array, physical_ticks = label_array_2d()
    pixel_ticks = {
        Axes.X: [2, 3, 4, 5, 6, 7, 8],
    }

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_label_array_and_ticks(
            label_array,
            pixel_ticks,
            physical_ticks,
            None
        )


def test_missing_physical_ticks_2d():
    """BinaryMaskCollection.from_label_array_and_ticks with some physical ticks missing."""
    label_array, physical_ticks_all = label_array_2d()

    for deleted_physical_ticks in physical_ticks_all.keys():
        physical_ticks = {
            coord: physical_ticks
            for coord, physical_ticks in physical_ticks_all.items()
            if coord != deleted_physical_ticks
        }
        with pytest.raises(ValueError):
            BinaryMaskCollection.from_label_array_and_ticks(
                label_array,
                None,
                physical_ticks,
                None
            )


def test_missing_physical_ticks_3d():
    """BinaryMaskCollection.from_label_array_and_ticks with some physical ticks missing."""
    label_array, physical_ticks_all = label_array_3d()

    for deleted_physical_ticks in physical_ticks_all.keys():
        physical_ticks = {
            coord: physical_ticks
            for coord, physical_ticks in physical_ticks_all.items()
            if coord != deleted_physical_ticks
        }
        with pytest.raises(ValueError):
            BinaryMaskCollection.from_label_array_and_ticks(
                label_array,
                None,
                physical_ticks,
                None
            )


def test_incorrectly_sized_physical_ticks():
    """BinaryMaskCollection.from_label_array_and_ticks with some physical ticks incorrectly
    sized."""
    label_image_array_2d, _ = label_array_2d()
    physical_ticks_2d = {Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
                         Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12]}

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_label_array_and_ticks(
            label_image_array_2d,
            None,
            physical_ticks_2d,
            None
        )

    label_image_array_3d, _ = label_array_3d()
    physical_ticks_3d = {
        Coordinates.Z: [0.0, 1.0],
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]
    }

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_label_array_and_ticks(
            label_image_array_3d,
            None,
            physical_ticks_3d,
            None
        )



from functools import partial
from typing import Optional, Union

import xarray as xr
from trackpy import bandpass

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number
from ._base import FilterAlgorithm
from .util import determine_axes_to_group_by


class Bandpass(FilterAlgorithm):
    """
    Convolve with a Gaussian to remove short-wavelength noise and subtract out long-wavelength
    variations, retaining features of intermediate scale. This implementation relies on
    scipy.ndimage.filters.gaussian_filter.

    This method is a thin wrapper around :code:`trackpy.Bandpass`:

    http://soft-matter.github.io/trackpy/v0.3.0/generated/trackpy.bandpass.html

    Parameters
    ----------
    lshort : float
        filter frequencies below this value
    llong : int
        filter frequencies above this odd integer value
    threshold : float
        zero any spots below this intensity value after background subtraction (default 0)
    truncate : float
        truncate the gaussian kernel, used by the gaussian filter, at this many standard
        deviations (default 4)
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) planes
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            lshort: Number,
            llong: int,
            threshold: Number = 0,
            truncate: Number = 4,
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:
        self.lshort = lshort
        self.llong = llong

        if threshold is None:
            raise ValueError("Threshold cannot be None, please pass a float or integer")

        self.threshold = threshold
        self.truncate = truncate
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"lshort": 1, "llong": 3, "threshold": 0.01}

    @staticmethod
    def _bandpass(
            image: xr.DataArray,
            lshort: Number, llong: int, threshold: Number, truncate: Number
    ) -> xr.DataArray:
        """Apply a bandpass filter to remove noise and background variation

        Parameters
        ----------
        image : xr.DataArray
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any spots below this intensity value after background subtraction (default 0)
        truncate : float
            truncate the gaussian kernel, used by the gaussian filter, at this many standard
            deviations

        Returns
        -------
        xr.DataArray :
            bandpassed image

        """
        bandpassed = bandpass(
            image, lshort=lshort, llong=llong, threshold=threshold,
            truncate=truncate
        )
        return bandpassed

    def run(
            self,
            stack: ImageStack,
            in_place: bool=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        bandpass_ = partial(
            self._bandpass,
            lshort=self.lshort, llong=self.llong, threshold=self.threshold, truncate=self.truncate
        )

        group_by = determine_axes_to_group_by(self.is_volume)

        result = stack.apply(
            bandpass_,
            group_by=group_by,
            in_place=in_place,
            n_processes=n_processes,
            level_method=self.level_method,
            verbose=verbose,
        )
        return result



from starfish.core.morphology.binary_mask.test.factories import binary_mask_collection_2d
from ..areafilter import AreaFilter


def test_empty_filter():
    input_mask_collection = binary_mask_collection_2d()
    output_mask_collection = AreaFilter(None, None).run(input_mask_collection)
    assert len(output_mask_collection) == len(input_mask_collection)

    for mask_num in range(len(input_mask_collection)):
        input_mask = input_mask_collection.uncropped_mask(mask_num)
        output_mask = output_mask_collection.uncropped_mask(mask_num)
        assert input_mask.equals(output_mask)


def test_min_area():
    input_mask_collection = binary_mask_collection_2d()
    output_mask_collection = AreaFilter(6, None).run(input_mask_collection)
    assert len(output_mask_collection) == 1


def test_max_area():
    input_mask_collection = binary_mask_collection_2d()
    output_mask_collection = AreaFilter(None, 5).run(input_mask_collection)
    assert len(output_mask_collection) == 1



"""To build experiments in-place, a few things must be done:

1. Call write_experiment_json with writer_contract=InplaceWriterContract().
2. The TileFetcher should return an instance of a InplaceTileFetcher.
"""


import abc
from pathlib import Path
from typing import Mapping, Optional

import numpy as np
from slicedimage import ImageFormat, Tile
from slicedimage.io import WriterContract

from starfish.core.types import Axes
from .providers import FetchedTile


class InplaceWriterContract(WriterContract):
    def tile_url_generator(self, tileset_url: str, tile: Tile, ext: str) -> str:
        return tile.provider.filepath.as_uri()

    def write_tile(
            self,
            tile_url: str,
            tile: Tile,
            tile_format: ImageFormat,
            backend_config: Optional[Mapping] = None,
    ) -> str:
        return tile.provider.sha256


class InplaceFetchedTile(FetchedTile):
    """Data formatters that operate in-place should return tiles that extend this class."""
    @property
    @abc.abstractmethod
    def filepath(self) -> Path:
        """Returns the path of the source tile."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def sha256(self):
        """Returns the sha256 checksum of the source tile."""
        raise NotImplementedError()

    def tile_data(self) -> np.ndarray:
        return np.zeros(shape=(self.shape[Axes.Y], self.shape[Axes.X]), dtype=np.float32)



from starfish import Experiment


def MERFISH(use_test_data: bool = False) -> Experiment:
    """
    Loads an experiment with a single field of view of MERFISH data derived from cultured U2-OS
    cells, published in the following manuscript:

    `<https://doi.org/10.1073/pnas.1612826113>`_

    The data consist of 16 images from 8 rounds, 2 channels, and a single z-plane. Each image is
    (2048, 2048) (y, x)

    Parameters
    ----------
    use_test_data: bool
        If True, returns a cropped Experiment where tiles are (205, 405) (y, x)

    Returns
    -------
    Experiment
    """
    if use_test_data:
        return Experiment.from_json(
            "https://d2nhj9g34unfro.cloudfront.net/20181005/MERFISH-TEST/experiment.json")
    return Experiment.from_json("https://d2nhj9g34unfro.cloudfront.net/browse/formatted/"
                                "MERFISH/20190511/experiment.json")


def allen_smFISH(use_test_data: bool = False) -> Experiment:
    """
    Loads an experiment with a single round from a single field of view of sequential smFISH data
    taken from mouse primary visual cortex. These data are unpublished, and were kindly
    contributed by the Allen Institute for Brain Science as a part of the SpaceTx consortium
    project.

    The data consist of 102 images from 1 round, 3 channels, and 33 z-planes. Each image is
    (2048, 2048) (y, x). There are no test data.

    Parameters
    ----------
    use_test_data : bool
        Not used.

    Returns
    -------
    Experiment
    """
    if use_test_data:
        return Experiment.from_json(
            "https://d2nhj9g34unfro.cloudfront.net/20181005/allen_smFISH/experiment.json")
    return Experiment.from_json(
        "https://d26bikfyahveg8.cloudfront.net/smFISH/mouse/formatted_with_DAPI/experiment.json")


def DARTFISH(use_test_data: bool = False) -> Experiment:
    """
    Loads an experiment with a single field of view from unpublished data generated with DARTFISH v1
    2017, produced by imaging human occipital cortex. These data were donated by the Zhang lab as
    part of the SpaceTx consortium project.

    The data consist of 18 images from 3 channels, 6 rounds, and 1 z-plane. Each image is (998, 998)
    in (y, x).

    Parameters
    ----------
    use_test_data : bool
        If True, returns a cropped Experiment where tiles are (170, 290) in (y, x)

    Returns
    -------
    Experiment
    """
    if use_test_data:
        return Experiment.from_json(
            "https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH-TEST/experiment.json")
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH/experiment.json")


def ISS(use_test_data: bool = False) -> Experiment:
    """
    Loads an experiment containing a single field of view of In-Situ Sequencing data created by
    imaging human breast cancer tissue. The complete dataset is published:

    `<https://doi.org/10.1038/nmeth.2563>`_

    The data consist of 16 images from 4 channels, 4 rounds, and a single z-plane. Each image is
    (1044, 1390) in (y, x)

    Parameters
    ----------
    use_test_data : bool
        If True, returns a cropped Experiment where images are (140, 200) in (y, x)

    Returns
    -------
    Experiment
    """
    if use_test_data:
        return Experiment.from_json(
            "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/experiment.json")
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/iss/20190506/experiment.json")


def osmFISH(use_test_data: bool = False) -> Experiment:
    """
    Loads an experiment containing 3 fields of view of osmFISH data generated by imaging mouse
    sematosensory cortex. The complete dataset is published:

    `<https://doi.org/10.1038/s41592-018-0175-z>`_

    The data consist of 1755 images per field of view, taken over 13 rounds, 3 channels, and 45
    z-planes. Each image is (2048, 2048) in (y, x)

    Parameters
    ----------
    use_test_data : bool
        If True, return a single round from one field of view, suitable for testing. The images are
        still (2048, 2048) in (y, x)

    Returns
    -------
    Experiment
    """
    if use_test_data:
        return Experiment.from_json(
            "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181217/osmFISH/"
            "experiment.json"
        )
    return Experiment.from_json(
        "https://d26bikfyahveg8.cloudfront.net/osmFISH/formatted/20190626/experiment.json")


def BaristaSeq(use_test_data: bool = False) -> Experiment:
    """Loads a BaristaSeq dataset generated from mouse visual cortex. The extracted field of view
    comes from an internal layer of V1 (range: 2-5). These data are published here:

    `<https://doi.org/10.1093/nar/gkx1206>`_

    The data consist of 204 images acquired from 3 rounds, 4 channels, and 17 z-planes. Each image
    is (1000, 800) in (y, x)

    Parameters
    ----------
    use_test_data : bool
        Not used.

    Returns
    -------
    Experiment
    """
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181028/"
        "BaristaSeq/cropped_formatted/experiment.json"
    )


def ImagingMassCytometry(use_test_data: bool = False) -> Experiment:
    """
    Loads an Imaging Mass Cytometry dataset donated to starfish by Denis Shapiro. This dataset
    consists of 52 fields of view.

    Each field of view consists of 19 images acquired from 19 channels, 1 round, and 1 z-plane.
    Individual images have variable sizes, ranging from 400-500 pixels for each of (y, x).

    Parameters
    ----------
    use_test_data:
        Not used.

    Returns
    -------
    Experiment
    """
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181023/"
        "imaging_cytof/BodenmillerBreastCancerSamples/experiment.json"
    )


def SeqFISH(use_test_data: bool = False) -> Experiment:
    """
    Loads a single field of view from a SeqFISH dataset generated from cultured mES cells. These
    data are published:

    `<https://doi.org/10.1016/j.cell.2018.05.035>`_

    This dataset contains 1140 images acquired across 12 channels, 4 imaging rounds, and 29
    z-planes. Each image is (2048, 2048) in (y, x)

    Parameters
    ----------
    use_test_data : bool
        If true, return a cropped Experiment where images are (280, 280) in (y, x)

    Returns
    -------
    Experiment
    """
    suffix = "-TEST" if use_test_data else ""
    url = (
        f"https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181211/"
        f"seqfish{suffix}/experiment.json"
    )
    return Experiment.from_json(url)


def STARmap(use_test_data: bool = False) -> Experiment:
    """
    Loads a STARmap field of view generated from mouse primary visual cortex. These data are
    published:

    `<https://doi.org/10.1126/science.aat5691>`_

    This dataset contains 696 images acquired across 6 rounds, 4 channels, and 29 z-planes. Each
    image is (1024, 1024) in (y, x)

    Parameters
    ----------
    use_test_data : bool
        Not used.

    Returns
    -------
    Experiment
    """
    url = (
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20190309/"
        "starmap/experiment.json"
    )
    return Experiment.from_json(url)



import numpy as np
from skimage.feature import register_translation
from skimage.transform._geometric import SimilarityTransform

from starfish.core.image._registration.transforms_list import TransformsList
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes, TransformType
from ._base import LearnTransformAlgorithm


class Translation(LearnTransformAlgorithm):
    """
    Iterate over the given axes of an ImageStack and learn the translation transform
    based off the reference_stack passed into :py:class:`Translation`'s constructor.
    Only supports 2d data.

    Parameters
    ----------
    axes : Axes
        The axes {r, ch, zplane} to iterate over
    reference_stack : ImageStack
        The target image used in :py:func:`skimage.feature.register_translation`
    upsampling : int
        upsampling factor (default=1). See
        http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.register_translation
        for an explanation of this parameter. in brief, this parameter determines the resolution of
        the registration. A value of 1 represents pixel resolution, a value of 10 is 1/10th of
        a pixel, a value of 300 is 1/300th of a pixel, and so on.
    """

    def __init__(self, reference_stack: ImageStack, axes: Axes, upsampling: int=1):

        self.upsampling = upsampling
        self.axes = axes
        self.reference_stack = reference_stack

    def run(self, stack: ImageStack, verbose: bool=False, *args) -> TransformsList:
        """
        Iterate over the given axes of an ImageStack and learn the translation transform
        based off the reference_stack passed into :py:class:`Translation`'s constructor.
        Only supports 2d data.

        Parameters
        ----------
        stack : ImageStack
            Stack to calculate the transforms on.
        verbose : bool
            if True, report on transformation progress (default = False)

        Returns
        -------
        List[Tuple[Mapping[Axes, int], SimilarityTransform]] :
            A list of tuples containing axes of the Imagestack and associated
            transform to apply.
        """

        transforms = TransformsList()
        reference_image = np.squeeze(self.reference_stack.xarray)
        for a in stack.axis_labels(self.axes):
            target_image = np.squeeze(stack.sel({self.axes: a}).xarray)
            if len(target_image.shape) != 2:
                raise ValueError(
                    f"Only axes: {self.axes.value} can have a length > 1, "
                    f"please use the MaxProj filter."
                )

            shift, error, phasediff = register_translation(src_image=target_image,
                                                           target_image=reference_image,
                                                           upsample_factor=self.upsampling)
            if verbose:
                print(f"For {self.axes}: {a}, Shift: {shift}, Error: {error}")
            selectors = {self.axes: a}
            # reverse shift because SimilarityTransform stores in y,x format
            shift = shift[::-1]
            transforms.append(selectors,
                              TransformType.SIMILARITY,
                              SimilarityTransform(translation=shift))

        return transforms



"""
This module describes the contracts to provide data to the experiment builder.
"""

from typing import Mapping, Union

import numpy as np

from starfish.core.types import Axes, Coordinates, CoordinateValue


class FetchedTile:
    """
    This is the contract for providing the data for constructing a :class:`slicedimage.Tile`.
    """
    def __init__(self, *args, **kwargs):
        pass

    @property
    def shape(self) -> Mapping[Axes, int]:
        """Return Tile shape.

        Returns
        -------
        Mapping[Axis, int]
            The shape of the tile, mapping from Axes to its size.
        """
        raise NotImplementedError()

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:
        """Return the tile's coordinates in the global coordinate space..

        Returns
        -------
        Mapping[Union[str, Coordinates], CoordinateValue]
            Maps from a coordinate type (e.g. 'x', 'y', or 'z') to its value or range.
        """
        raise NotImplementedError()

    @property
    def extras(self) -> dict:
        """Return the extras data associated with the tile.

        Returns
        -------
        Mapping[str, Any]
            Maps from a key to its value.
        """
        return {}

    def tile_data(self) -> np.ndarray:
        """Return the image data representing the tile.  The tile must be row-major.

        Returns
        -------
        ndarray :
            The image data
        """
        raise NotImplementedError()


class TileFetcher:
    """
    This is the contract for providing the image data for constructing a
    :class:`slicedimage.Collection`.
    """
    def get_tile(
            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:
        """
        Given fov_id, round_label, ch_label, and zplane_label, return an instance of a
        :class:`.FetchedImage` that can be queried for the image data.
        """
        raise NotImplementedError()



from typing import Optional, Union

import xarray as xr
from skimage.morphology import ball, disk, white_tophat

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels
from ._base import FilterAlgorithm
from .util import determine_axes_to_group_by


class WhiteTophat(FilterAlgorithm):
    """
    Performs "white top hat" filtering of an image to enhance spots. White top hat filtering
    finds spots that are both smaller and brighter than their surroundings by subtracting an
    estimate of the background produced by a binary opening of the image using a disk-shaped
    structuring element.

    Parameters
    ----------
    masking_radius : int
        radius of the morphological masking structure in pixels
    is_volume : int
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.

    Notes
    -----
    See https://en.wikipedia.org/wiki/Top-hat_transform for more information

    """

    def __init__(
            self,
            masking_radius: int,
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:
        self.masking_radius = masking_radius
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"masking_radius": 3}

    def _white_tophat(self, image: xr.DataArray) -> xr.DataArray:
        if self.is_volume:
            structuring_element = ball(self.masking_radius)
        else:
            structuring_element = disk(self.masking_radius)
        return white_tophat(image, selem=structuring_element)

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        result = stack.apply(
            self._white_tophat,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method
        )
        return result



import warnings
from typing import (
    Optional,
    Set,
    Union
)

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Axes, Clip, FunctionSource, FunctionSourceBundle, Levels
from ._base import FilterAlgorithm


class Map(FilterAlgorithm):
    """
    Map from input to output by applying a specified function to the input.  The output must have
    the same shape as the input.

    Parameters
    ----------
    func : Union[str, FunctionSourceBundle]
        Function to apply across the dimension(s) specified by ``dims``.

        If this value is a string, then the ``module`` parameter is consulted to determine which
        python package is used to find the function.  If ``module`` is not specified, then the
        default is :py:attr:`FunctionSource.np`.

        If this value is a ``FunctionSourceBundle``, then the python package and module name is
        obtained from the bundle.
    module : Optional[FunctionSource]
        Python module that serves as the source of the function.  It must be listed as one of the
        members of :py:class:`FunctionSource`.

        Currently, the supported FunctionSources are:
        - ``np``: the top-level package of numpy
        - ``scipy``: the top-level package of scipy

        This is being deprecated in favor of specifying the function as a ``FunctionSourceBundle``.
    in_place : bool
        Execute the operation in-place.  (default: False)
    group_by : Set[Axes]
        Axes to split the data along.  For example, splitting a 2D array (axes: X, Y; size: 3, 4)
        by X results in 3 calls to the method, each with arrays of size 4.  (default {Axes.ROUND,
        Axes.CH, Axes.ZPLANE})
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.

    Examples
    --------
    Applying a divide function.
        >>> from starfish.core.imagestack.test.factories.synthetic_stack import synthetic_stack
        >>> from starfish.image import Filter
        >>> from starfish.types import Axes
        >>> stack = synthetic_stack()
        >>> divider = Filter.Map("divide", 4)
        >>> by_four = divider.run(stack)

    See Also
    --------
    starfish.core.types.Axes

    """

    def __init__(
        self,
            func: Union[str, FunctionSourceBundle],
            *func_args,
            module: Optional[FunctionSource] = None,
            in_place: bool = False,
            group_by: Optional[Set[Union[Axes, str]]] = None,
            clip_method: Optional[Clip] = None,
            level_method: Optional[Levels] = None,
            **func_kwargs,
    ) -> None:
        if isinstance(func, str):
            if module is not None:
                warnings.warn(
                    f"The module parameter is being deprecated.  Use "
                    f"`func=FunctionSource.{module.name}{func} instead.",
                    DeprecationWarning)
            else:
                module = FunctionSource.np
            self.func = module(func)
        elif isinstance(func, FunctionSourceBundle):
            if module is not None:
                raise ValueError(
                    f"When passing in the function as a `FunctionSourceBundle`, module should not "
                    f"be set."
                )
            self.func = func
        self.in_place = in_place
        if group_by is None:
            group_by = {Axes.ROUND, Axes.CH, Axes.ZPLANE}
        self.group_by: Set[Axes] = {Axes(axis) for axis in group_by}
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)
        self.func_args = func_args
        self.func_kwargs = func_kwargs

    _DEFAULT_TESTING_PARAMETERS = {"dims": ['r'], "func": 'max'}

    def run(
            self,
            stack: ImageStack,
            *args,
    ) -> Optional[ImageStack]:
        """Map from input to output by applying a specified function to the input.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        Returns
        -------
        Optional[ImageStack] :
            If in-place is False, return the results of filter as a new stack.  Otherwise return
            None

        """

        # Apply the reducing function
        return stack.apply(
            self.func.resolve(),
            *self.func_args,
            group_by=self.group_by,
            in_place=self.in_place,
            level_method=self.level_method,
            **self.func_kwargs)



from functools import partial
from typing import Any, List, Mapping, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.ndimage import label
from skimage.feature import peak_local_max
from skimage.measure import regionprops
from sympy import Line, Point
from tqdm import tqdm

from starfish.core.config import StarfishConfig
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.spots.FindSpots import spot_finding_utils
from starfish.core.types import (
    Axes,
    Features,
    Number,
    PerImageSliceSpotResults,
    SpotAttributes,
    SpotFindingResults
)
from ._base import FindSpotsAlgorithm


class LocalMaxPeakFinder(FindSpotsAlgorithm):
    """
    2-dimensional local-max peak finder that wraps skimage.feature.peak_local_max

    Parameters
    ----------
    min_distance : int
        Minimum number of pixels separating peaks in a region of 2 * min_distance + 1
        (i.e. peaks are separated by at least min_distance). To find the maximum number of
        peaks, use min_distance=1.
    stringency : int
    min_obj_area : int
    max_obj_area : int
    threshold : Optional[Number]
    measurement_type : str, {'max', 'mean'}
        default 'max' calculates the maximum intensity inside the object
    min_num_spots_detected : int
        When fewer than this number of spots are detected, spot searching for higher threshold
        values. (default = 3)
    is_volume : bool
        Not supported. For 3d peak detection please use TrackpyLocalMaxPeakFinder.
        (default=False)
    verbose : bool
        If True, report the percentage completed during processing
        (default = False)

    Notes
    -----
    http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.peak_local_max
    """

    def __init__(
        self, min_distance: int, stringency: int, min_obj_area: int, max_obj_area: int,
        threshold: Optional[Number]=None, measurement_type: str='max',
        min_num_spots_detected: int=3, is_volume: bool=False, verbose: bool=True
    ) -> None:

        self.min_distance = min_distance
        self.stringency = stringency
        self.min_obj_area = min_obj_area
        self.max_obj_area = max_obj_area
        self.threshold = threshold
        self.min_num_spots_detected = min_num_spots_detected

        self.measurement_function = self._get_measurement_function(measurement_type)

        self.is_volume = is_volume
        self.verbose = verbose

    def _compute_num_spots_per_threshold(self, img: np.ndarray) -> Tuple[np.ndarray, List[int]]:
        """Computes the number of detected spots for each threshold

        Parameters
        ----------
        img : np.ndarray
            The image in which to count spots

        Returns
        -------
        np.ndarray :
            thresholds
        List[int] :
            spot counts
        """

        # thresholds to search over
        thresholds = np.linspace(img.min(), img.max(), num=100)

        # number of spots detected at each threshold
        spot_counts = []

        # where we stop our threshold search
        stop_threshold = None

        if self.verbose and StarfishConfig().verbose:
            threshold_iter = tqdm(thresholds)
            print('Determining optimal threshold ...')
        else:
            threshold_iter = thresholds

        for stop_index, threshold in enumerate(threshold_iter):
            spots = peak_local_max(
                img,
                min_distance=self.min_distance,
                threshold_abs=threshold,
                exclude_border=False,
                indices=True,
                num_peaks=np.inf,
                footprint=None,
                labels=None
            )

            # stop spot finding when the number of detected spots falls below min_num_spots_detected
            if len(spots) <= self.min_num_spots_detected:
                stop_threshold = threshold
                if self.verbose:
                    print(f'Stopping early at threshold={threshold}. Number of spots fell below: '
                          f'{self.min_num_spots_detected}')
                break
            else:
                spot_counts.append(len(spots))

        if stop_threshold is None:
            stop_threshold = thresholds.max()

        if len(thresholds > 1):
            thresholds = thresholds[:stop_index]
            spot_counts = spot_counts[:stop_index]

        return thresholds, spot_counts

    def _select_optimal_threshold(self, thresholds: np.ndarray, spot_counts: List[int]) -> float:
        # calculate the gradient of the number of spots
        grad = np.gradient(spot_counts)
        optimal_threshold_index = np.argmin(grad)

        # only consider thresholds > than optimal threshold
        thresholds = thresholds[optimal_threshold_index:]
        grad = grad[optimal_threshold_index:]

        # if all else fails, return 0.
        selected_thr = 0

        if len(thresholds) > 1:

            distances = []

            # create a line whose end points are the threshold and and corresponding gradient value
            # for spot_counts corresponding to the threshold
            start_point = Point(thresholds[0], grad[0])
            end_point = Point(thresholds[-1], grad[-1])
            line = Line(start_point, end_point)

            # calculate the distance between all points and the line
            for k in range(len(thresholds)):
                p = Point(thresholds[k], grad[k])
                dst = line.distance(p)
                distances.append(dst.evalf())

            # remove the end points
            thresholds = thresholds[1:-1]
            distances = distances[1:-1]

            # select the threshold that has the maximum distance from the line
            # if stringency is passed, select a threshold that is n steps higher, where n is the
            # value of stringency
            if distances:
                thr_idx = np.argmax(np.array(distances))

                if thr_idx + self.stringency < len(thresholds):
                    selected_thr = thresholds[thr_idx + self.stringency]
                else:
                    selected_thr = thresholds[thr_idx]

        return selected_thr

    def _compute_threshold(
            self, img: np.ndarray) -> Tuple[float, Optional[np.ndarray], Optional[List[int]]]:
        """Finds spots on a number of thresholds then selects and returns the optimal threshold

        Parameters
        ----------
        img: np.ndarray
            data array in which spots should be detected and over which to compute different
            intensity thresholds

        Returns
        -------
        np.ndarray :
            The intensity threshold
        """
        img = np.asarray(img)
        thresholds, spot_counts = self._compute_num_spots_per_threshold(img)
        if len(spot_counts) == 0:
            # this only happens when we never find more spots than `self.min_num_spots_detected`
            return img.min(), None, None
        return self._select_optimal_threshold(thresholds, spot_counts), thresholds, spot_counts

    def image_to_spots(
            self, data_image: np.ndarray
    ) -> PerImageSliceSpotResults:
        """measure attributes of spots detected by binarizing the image using the selected threshold

        Parameters
        ----------
        data_image : np.ndarray
            image containing spots to be detected

        Returns
        -------
        PerImageSpotResults :
            includes a SpotAttributes DataFrame of metadata containing the coordinates, intensity
            and radius of each spot, as well as any extra information collected during spot finding.
        """

        optimal_threshold, thresholds, spot_counts = self._compute_threshold(data_image)

        data_image = np.asarray(data_image)

        # identify each spot's size by binarizing and calculating regionprops
        masked_image = data_image[:, :] > optimal_threshold
        labels = label(masked_image)[0]
        spot_props = regionprops(np.squeeze(labels))

        # mask spots whose areas are too small or too large
        for spot_prop in spot_props:
            if spot_prop.area < self.min_obj_area or spot_prop.area > self.max_obj_area:
                masked_image[0, spot_prop.coords[:, 0], spot_prop.coords[:, 1]] = 0

        # store re-calculated regionprops and labels based on the area-masked image
        labels = label(masked_image)[0]

        if self.verbose:
            print('computing final spots ...')

        spot_coords = peak_local_max(
            data_image,
            min_distance=self.min_distance,
            threshold_abs=optimal_threshold,
            exclude_border=False,
            indices=True,
            num_peaks=np.inf,
            footprint=None,
            labels=labels
        )

        res = {Axes.X.value: spot_coords[:, 2],
               Axes.Y.value: spot_coords[:, 1],
               Axes.ZPLANE.value: spot_coords[:, 0],
               Features.SPOT_RADIUS: 1,
               Features.SPOT_ID: np.arange(spot_coords.shape[0]),
               Features.INTENSITY: data_image[spot_coords[:, 0],
                                              spot_coords[:, 1],
                                              spot_coords[:, 2]],
               }
        extras: Mapping[str, Any] = {
            "threshold": optimal_threshold,
            "thresholds": thresholds,
            "spot_counts": spot_counts
        }

        return PerImageSliceSpotResults(spot_attrs=SpotAttributes(pd.DataFrame(res)), extras=extras)

    def run(
            self,
            image_stack: ImageStack,
            reference_image: Optional[ImageStack] = None,
            n_processes: Optional[int] = None,
            *args,
            **kwargs
    ) -> SpotFindingResults:
        """
        Find spots in the given ImageStack using a gaussian blob finding algorithm.
        If a reference image is provided the spots will be detected there then measured
        across all rounds and channels in the corresponding ImageStack. If a reference_image
        is not provided spots will be detected _independently_ in each channel. This assumes
        a non-multiplex imaging experiment, as only one (ch, round) will be measured for each spot.

        Parameters
        ----------
        image_stack : ImageStack
            ImageStack where we find the spots in.
        reference_image : Optional[ImageStack]
            (Optional) a reference image. If provided, spots will be found in this image, and then
            the locations that correspond to these spots will be measured across each channel.
        n_processes : Optional[int] = None,
            Number of processes to devote to spot finding.
        """
        spot_finding_method = partial(self.image_to_spots, *args, **kwargs)
        if reference_image:
            data_image = reference_image._squeezed_numpy(*{Axes.ROUND, Axes.CH})
            reference_spots = spot_finding_method(data_image)
            results = spot_finding_utils.measure_intensities_at_spot_locations_across_imagestack(
                data_image=image_stack,
                reference_spots=reference_spots,
                measurement_function=self.measurement_function)
        else:
            spot_attributes_list = image_stack.transform(
                func=spot_finding_method,
                group_by={Axes.ROUND, Axes.CH},
                n_processes=n_processes
            )
            results = SpotFindingResults(imagestack_coords=image_stack.xarray.coords,
                                         log=image_stack.log,
                                         spot_attributes_list=spot_attributes_list)
        return results



from functools import partial
from typing import Optional, Set

import numpy as np
import xarray as xr

from starfish.core.compat import match_histograms
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes
from starfish.core.util import enum
from ._base import FilterAlgorithm


class MatchHistograms(FilterAlgorithm):
    """
    Normalize data by matching distributions of each tile or volume to a reference volume

    Chunks sharing the same values for axes specified by group_by will be quantile
    normalized such that their intensity values are identically distributed. The reference
    distribution is calculated by sorting the intensities in each chunk and averaging across
    chunks.

    For example, if group_by={Axes.CH, Axes.ROUND} each (z, y, x) volume will be linearized,
    the intensities will be sorted, and averaged across {Axes.CH, Axes.ROUND}, equalizing
    the intensity distribution of each (round, channel) volume.

    Setting group_by={Axes.CH} would carry out the same approach, but the result would _retain_
    variation across channel.

    Parameters
    ----------
    group_by : Set[Axes]
        The Axes to group by.
    """

    def __init__(
            self, group_by: Set[Axes]
    ) -> None:

        self.group_by = group_by

    _DEFAULT_TESTING_PARAMETERS = {"group_by": {Axes.CH, Axes.ROUND}}

    def _compute_reference_distribution(self, data: ImageStack) -> xr.DataArray:
        """compute the average reference distribution across the ImageStack"""
        chunk_key = enum.harmonize(data.shape.keys() - self.group_by)
        sort_key = enum.harmonize(self.group_by)

        # stack up the array
        stacked = data.xarray.stack(chunk_key=chunk_key)
        stacked = stacked.stack(sort_key=sort_key)

        sorted_stacked = stacked.groupby("sort_key").apply(np.sort)
        reference = sorted_stacked.mean("sort_key")
        reference = reference.unstack("chunk_key")
        return reference

    @staticmethod
    def _match_histograms(
        image: xr.DataArray, reference: np.ndarray
    ) -> np.ndarray:
        """
        matches the intensity distribution of image to reference

        Parameters
        ----------
        image, reference : xr.DataArray
            3-d image data

        Returns
        -------
        np.ndarray :
            image, with intensities matched to reference
        """
        image = np.asarray(image)
        reference = np.asarray(reference)
        return match_histograms(image, reference=reference)

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        if verbose:
            print("Calculating reference distribution...")
        reference_image = self._compute_reference_distribution(stack)
        apply_function = partial(self._match_histograms, reference=reference_image)
        result = stack.apply(
            apply_function,
            group_by=self.group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result



from copy import deepcopy
from typing import Optional, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels
from starfish.core.util.levels import levels
from ._base import FilterAlgorithm


class ElementWiseMultiply(FilterAlgorithm):
    """
    Perform element-wise multiplication on the image tensor. This is useful for
    performing operations such as image normalization or field flatness correction

    Parameters
    ----------
    mult_mat : xr.DataArray
        the image is element-wise multiplied with this array
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
    """

    def __init__(
            self,
            mult_array: xr.DataArray,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None,
    ) -> None:

        self.mult_array = mult_array
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)
        if self.level_method in (Levels.SCALE_BY_CHUNK, Levels.SCALE_SATURATED_BY_CHUNK):
            raise ValueError("`scale_by_chunk` is not a valid clip_method for ElementWiseMultiply")

    _DEFAULT_TESTING_PARAMETERS = {
        "mult_array": xr.DataArray(
            np.array([[[[[1]]], [[[0.5]]]]]),
            dims=('r', 'c', 'z', 'y', 'x')
        )
    }

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        # Align the axes of the multipliers with ImageStack
        mult_array_aligned: np.ndarray = self.mult_array.transpose(*stack.xarray.dims).values
        if not in_place:
            stack = deepcopy(stack)
            self.run(stack, in_place=True)
            return stack

        stack.xarray.values *= mult_array_aligned
        if self.level_method == Levels.CLIP:
            stack.xarray.values = levels(stack.xarray.values)
        elif self.level_method == Levels.SCALE_BY_IMAGE:
            stack.xarray.values = levels(stack.xarray.values, rescale=True)
        elif self.level_method == Levels.SCALE_SATURATED_BY_IMAGE:
            stack.xarray.values = levels(stack.xarray.values, rescale_saturated=True)
        else:
            raise ValueError(
                f"Unknown level method {self.level_method}. See starfish.types.Levels for valid "
                f"options")
        return None



from typing import Callable, Hashable, Mapping, Optional

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, Coordinates, Features, SpotFindingResults


def transfer_physical_coords_to_intensity_table(
        *,
        intensity_table: IntensityTable,
        image_stack: Optional[ImageStack] = None,
        spots: Optional[SpotFindingResults] = None,
) -> IntensityTable:
    """
    Transfers physical coordinates from either an Imagestack or SpotFindingResults to an
    intensity table

    1. Creates three new coords on the intensity table (xc, yc, zc)
    2. For every spot:
        - Get pixel x,y values
        - Calculate the physical x,y values
        - Assign those values to the coords arrays for this spot
    """

    pairs = (
        (Axes.X.value, Coordinates.X.value),
        (Axes.Y.value, Coordinates.Y.value),
        (Axes.ZPLANE.value, Coordinates.Z.value)
    )
    if bool(image_stack) == bool(spots):
        raise ValueError("Must provide either SpotFindingResults or ImageStack to calculate "
                         "coordinates from, and not both.")

    coord_ranges: Mapping[Hashable, xr.DataArray]
    if image_stack is not None:
        coord_ranges = {
            Axes.X.value: image_stack.xarray[Coordinates.X.value],
            Axes.Y.value: image_stack.xarray[Coordinates.Y.value],
            Axes.ZPLANE.value: image_stack.xarray[Coordinates.Z.value]
        }
    elif spots is not None:
        coord_ranges = spots.physical_coord_ranges

    # make sure the intensity table gets empty metadata if there are no intensities
    if intensity_table.sizes[Features.AXIS] == 0:
        for axis, coord in pairs:
            intensity_table[coord] = xr.DataArray(np.zeros((0)), dims=Features.AXIS)
        return intensity_table

    for axis, coord in pairs:
        pixels: xr.DataArray = coord_ranges[axis]
        intensity_table_pixel_offsets: np.ndarray = intensity_table[axis].values

        # can't interpolate if the axis size == 1, so just select in that case.
        coordinate_fetcher: Callable
        if len(pixels) == 1:
            coordinate_fetcher = pixels.sel
        else:
            coordinate_fetcher = pixels.interp
        coordinates = coordinate_fetcher({axis: intensity_table_pixel_offsets})[coord]

        intensity_table[coord] = xr.DataArray(coordinates.values, dims='features')

    return intensity_table



import numpy as np
import pytest

from starfish import ImageStack
from starfish.core.spots.DecodeSpots import trace_builders
from starfish.core.spots.FindSpots import BlobDetector
from starfish.core.spots.FindSpots import FindSpotsAlgorithm
from starfish.core.test.factories import (
    two_spot_informative_blank_coded_data_factory,
    two_spot_one_hot_coded_data_factory,
    two_spot_sparse_coded_data_factory,
)
from starfish.types import Axes, Features, FunctionSource


# verify all spot finders handle different coding types
_, ONE_HOT_IMAGESTACK, ONE_HOT_MAX_INTENSITY = two_spot_one_hot_coded_data_factory()
_, SPARSE_IMAGESTACK, SPARSE_MAX_INTENSITY = two_spot_sparse_coded_data_factory()
_, BLANK_IMAGESTACK, BLANK_MAX_INTENSITY = two_spot_informative_blank_coded_data_factory()

# make sure that all spot finders handle empty arrays
EMPTY_IMAGESTACK = ImageStack.from_numpy(np.zeros((4, 2, 10, 100, 100), dtype=np.float32))


def simple_gaussian_spot_detector() -> BlobDetector:
    """create a basic gaussian spot detector"""
    return BlobDetector(
        min_sigma=1,
        max_sigma=4,
        num_sigma=5,
        threshold=0,
        measurement_type='max')


# initialize spot detectors
gaussian_spot_detector = simple_gaussian_spot_detector()


# test parameterization
test_parameters = (
    'data_stack, spot_detector, max_intensity',
    [
        (ONE_HOT_IMAGESTACK, gaussian_spot_detector, ONE_HOT_MAX_INTENSITY),
        (SPARSE_IMAGESTACK, gaussian_spot_detector, SPARSE_MAX_INTENSITY),
        (BLANK_IMAGESTACK, gaussian_spot_detector, BLANK_MAX_INTENSITY),
    ]
)


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_with_reference_image_exact_match(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
        max_intensity: float,
):
    """This testing method uses a reference image to identify spot locations then builds traces
    using the exact_match strategy. This represents a workflow common in a multiplexed assays.
    Each method should detect 2 total spots in the max projected reference image then group them
    into 2 distinct spot traces across the ImageStack. """

    reference_image = data_stack.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spots = spot_detector.run(image_stack=data_stack, reference_image=reference_image)
    intensity_table = trace_builders.build_spot_traces_exact_match(spots)
    assert intensity_table.sizes[Features.AXIS] == 2, "wrong number of spots traces detected"
    expected = [max_intensity * 2, max_intensity * 2]
    assert np.allclose(intensity_table.sum((Axes.ROUND, Axes.CH)).values, expected), \
        "wrong spot intensities detected"

    # verify this execution strategy produces an empty intensitytable when called with a blank image
    reference_image = EMPTY_IMAGESTACK.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spots = spot_detector.run(image_stack=EMPTY_IMAGESTACK, reference_image=reference_image)
    empty_intensity_table = trace_builders.build_spot_traces_exact_match(spots)
    assert empty_intensity_table.sizes[Features.AXIS] == 0


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_no_reference_image_exact_match(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
        max_intensity: float,
):
    """
    This testing method does not provide a reference image, and should therefore check for spots
    in each (round, ch) combination in sequence. It then uses the exact_match strategy to build
    spot traces. Since the spot finding only finds 1 spot in the first tile, the trace builder
    only builds one trace. This workflow doesn't really make real world sense but we're testing it
    anyway.
    """
    spots = spot_detector.run(image_stack=data_stack)
    intensity_table = trace_builders.build_spot_traces_exact_match(spots)
    assert intensity_table.sizes[Features.AXIS] == 1, "wrong number of spots traces detected"

    spots = spot_detector.run(image_stack=EMPTY_IMAGESTACK)
    empty_intensity_table = trace_builders.build_spot_traces_exact_match(spots)
    assert empty_intensity_table.sizes[Features.AXIS] == 0


@pytest.mark.parametrize(*test_parameters)
def test_spot_finding_no_reference_image_sequential(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
        max_intensity: float,
):
    """
    This testing method does not provide a reference image, and should therefore check for spots
    in each (round, ch) combination in sequence. With the given input, it should detect 4 spots.
    It then uses the sequential trace building strategy which should treat each spot as it's own
    unique trace totalling 4 traces.
    """
    spots = spot_detector.run(image_stack=data_stack)
    intensity_table = trace_builders.build_traces_sequential(spots)
    assert intensity_table.sizes[Features.AXIS] == 4, "wrong number of spot traces detected"
    expected = [max_intensity] * 4
    assert np.allclose(intensity_table.sum((Axes.ROUND, Axes.CH)).values, expected), \
        "wrong spot intensities detected"

    spots = spot_detector.run(image_stack=EMPTY_IMAGESTACK)
    empty_intensity_table = trace_builders.build_traces_sequential(spots)
    assert empty_intensity_table.sizes[Features.AXIS] == 0


@pytest.mark.parametrize(*test_parameters)
def test_spot_finding_reference_image_sequential(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
        max_intensity: float,
):
    """
    This testing method uses a reference image to identify spot locations then builds traces
    using the sequential strategy. It finds 2 spots in the max projected image, then measures the
    two spots on each tile totally 2*num_chs*num_rounds spots. When building spot traces it treats
    each spot as it's own trace totally 8 traces. This workflow doesn't really make sense but
    we're testing it anyway.
    """

    reference_image = data_stack.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spots = spot_detector.run(image_stack=data_stack, reference_image=reference_image)
    intensity_table = trace_builders.build_traces_sequential(spots)
    expected_num_traces = (2 * data_stack.num_chs * data_stack.num_rounds)
    assert intensity_table.sizes[Features.AXIS] == expected_num_traces, "wrong number of " \
                                                                        "spots traces detected"

    reference_image = EMPTY_IMAGESTACK.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spots = spot_detector.run(image_stack=EMPTY_IMAGESTACK, reference_image=reference_image)
    empty_intensity_table = trace_builders.build_traces_sequential(spots)
    assert empty_intensity_table.sizes[Features.AXIS] == 0



import os
import tempfile
from collections import OrderedDict

import numpy as np

from starfish.core.imagestack import indexing_utils as iu
from starfish.core.types import Axes, Coordinates, PhysicalCoordinateTypes
from .factories import imagestack_with_coords_factory, synthetic_stack
from ..imagestack import ImageStack


def test_imagestack_indexing():
    """Tests indexing on an Imagestack with a shape (5, 5, 15, 200, 200)
        steps:
        1.) stack.sel(indexers)
        2.) assert new shape of stack is what we expect

    """

    stack = synthetic_stack(num_round=5, num_ch=5, num_z=15, tile_height=200, tile_width=200)

    # index on range of rounds and single ch and Z
    indexed = stack.sel({Axes.ROUND: (1, None), Axes.CH: 0, Axes.ZPLANE: 0})
    expected_shape = OrderedDict([(Axes.ROUND, 4), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 200), (Axes.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round ch and z
    indexed = stack.sel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 0})
    expected_shape = OrderedDict([(Axes.ROUND, 1), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 200), (Axes.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round and range of ch
    indexed = stack.sel({Axes.ROUND: 1, Axes.CH: (3, None)})
    expected_shape = OrderedDict(
        [(Axes.ROUND, 1), (Axes.CH, 2), (Axes.ZPLANE, 15), (Axes.Y, 200), (Axes.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round and range of ch and Z
    indexed = stack.sel({Axes.ROUND: 1, Axes.CH: (None, 3), Axes.ZPLANE: (7, None)})
    expected_shape = OrderedDict(
        [(Axes.ROUND, 1), (Axes.CH, 4), (Axes.ZPLANE, 8), (Axes.Y, 200), (Axes.X, 200)])
    assert indexed.shape == expected_shape

    # index on first half of X and single value of Y
    indexed_stack = stack.sel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 1,
                               Axes.Y: 100, Axes.X: (None, 100)})
    expected_shape = OrderedDict([(Axes.ROUND, 1), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 1), (Axes.X, 100)])
    assert indexed_stack.shape == expected_shape

    # index on first half of X and Y
    indexed_stack = stack.sel({Axes.Y: (None, 100), Axes.X: (None, 100)})

    expected_shape = OrderedDict([(Axes.ROUND, 5), (Axes.CH, 5),
                                  (Axes.ZPLANE, 15), (Axes.Y, 100), (Axes.X, 100)])
    assert indexed_stack.shape == expected_shape

    # index on single x and y
    indexed_stack = stack.sel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 1,
                               Axes.Y: 100, Axes.X: 150})
    expected_shape = OrderedDict([(Axes.ROUND, 1), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 1), (Axes.X, 1)])
    assert indexed_stack.shape == expected_shape

    # Negative indexing
    indexed_stack = stack.sel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 1,
                               Axes.Y: (None, -10), Axes.X: (None, -10)})
    expected_shape = OrderedDict([(Axes.ROUND, 1), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 190), (Axes.X, 190)])
    assert indexed_stack.shape == expected_shape


X_COORDS = 1, 2
Y_COORDS = 4, 6
Z_COORDS = 1, 3


def test_find_nearest():
    """
    Set up ImageStack with physical coordinates:
        x_coords = [1. 1.11111111 1.22222222 1.33333333 1.44444444 1.55555556
        1.66666667 1.77777778 1.88888889 2.]

        y_coords = [4. 4.22222222 4.44444444 4.66666667 4.88888889 5.11111111
        5.33333333 5.55555556 5.77777778 6. ]

     Test that find_nearest() finds the correct corresponding positional index values
    """
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 10), (Axes.X, 10)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)
    assert iu.find_nearest(stack.xarray[Coordinates.X.value], 1.2) == 2
    assert iu.find_nearest(stack.xarray[Coordinates.X.value], 1.5) == 4
    assert iu.find_nearest(stack.xarray[Coordinates.X.value], (1.2, 1.5)) == (2, 4)

    assert iu.find_nearest(stack.xarray[Coordinates.Y.value], 4) == 0
    assert iu.find_nearest(stack.xarray[Coordinates.Y.value], 5.1) == 5
    assert iu.find_nearest(stack.xarray[Coordinates.Y.value], (4, 5.1)) == (0, 5)

    # assert values outside the range are given min/max of array
    assert iu.find_nearest(stack.xarray[Coordinates.X.value], 5) == 9
    assert iu.find_nearest(stack.xarray[Coordinates.X.value], -5) == 0


def test_convert_coords_to_indices():
    """
    Set up ImageStack with physical coordinates:
        x_coords = [1. 1.11111111 1.22222222 1.33333333 1.44444444 1.55555556
        1.66666667 1.77777778 1.88888889 2.]

        y_coords = [4. 4.22222222 4.44444444 4.66666667 4.88888889 5.11111111
        5.33333333 5.55555556 5.77777778 6. ]

    Test that convert_coords_to_indices() correctly converts Coordinate indices
    to their corresponding positional indices
    """
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 10), (Axes.X, 10)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)
    coordinate_indices = {Coordinates.X: (1.2, 1.5), Coordinates.Y: (4, 5.1)}
    positional_indices = iu.convert_coords_to_indices(stack.xarray, coordinate_indices)

    assert positional_indices[Axes.X] == iu.find_nearest(
        stack.xarray[Coordinates.X.value], (1.2, 1.5))
    assert positional_indices[Axes.Y] == iu.find_nearest(
        stack.xarray[Coordinates.Y.value], (4, 5.1))


def test_sel_by_physical_coords():
    """
    Set up ImageStack with physical coordinates:
        x_coords = [1. 1.11111111 1.22222222 1.33333333 1.44444444 1.55555556
        1.66666667 1.77777778 1.88888889 2.]

        y_coords = [4. 4.22222222 4.44444444 4.66666667 4.88888889 5.11111111
        5.33333333 5.55555556 5.77777778 6. ]

    Test that sel_by_physical_coords() correctly indexes the imagestack by the
    corresponding positional indexers
    """
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 10), (Axes.X, 10)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)

    indexed_stack_by_coords = stack.sel_by_physical_coords({Coordinates.X: (1.2, 1.5),
                                                            Coordinates.Y: (4, 5.1)})
    indexed_stack_by_pos = stack.sel({Axes.X: (2, 4), Axes.Y: (0, 5)})

    # assert that the resulting xarrays are the same
    assert indexed_stack_by_coords.xarray.equals(indexed_stack_by_pos.xarray)


def test_sel_by_physical_and_axes():
    """
       Set up ImageStack with physical coordinates:
           x_coords = [1. 1.11111111 1.22222222 1.33333333 1.44444444 1.55555556
           1.66666667 1.77777778 1.88888889 2.]

           y_coords = [4. 4.22222222 4.44444444 4.66666667 4.88888889 5.11111111
           5.33333333 5.55555556 5.77777778 6. ]

       Test that sel_by_physical_coords() correctly indexes the imagestack by the
       corresponding positional indexers
       """
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 10), (Axes.X, 10)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)
    indexed_stack_by_coords = stack.sel_by_physical_coords({Coordinates.X: (1.2, 1.5),
                                                            Coordinates.Y: (4, 5.1)})
    indexed_stack = indexed_stack_by_coords.sel({Axes.ROUND: 2, Axes.CH: 1, Axes.ZPLANE: 0})
    expected_shape = OrderedDict([(Axes.ROUND, 1), (Axes.CH, 1),
                                  (Axes.ZPLANE, 1), (Axes.Y, 5), (Axes.X, 2)])
    assert indexed_stack.shape == expected_shape


def test_nonindexed_dimensions_restored():
    """When the selection removes a dimension, xarray.expand_dims does not expand the non-indexed
    dimensions that were removed.  For example, if one selects only a single zplane, it reduce the z
    physical coordinate to a coordinate scalar, and not an array of size 1.  This verifies that the
    workaround we introduced to restore the dependent axes's labels is in place.
    """
    stack = synthetic_stack(num_round=5, num_ch=5, num_z=15, tile_height=200, tile_width=200)

    for selector in (
            {Axes.ROUND: 0, Axes.CH: 2, Axes.ZPLANE: 5},
            {Axes.ROUND: (0, 3), Axes.CH: 2, Axes.ZPLANE: 5},
            {Axes.CH: (None, 3), Axes.ZPLANE: 5},
    ):
        sel_xarray = stack.sel(selector).xarray

        # when the selection removes a dimension (e.g., only select a single z plane)
        for primary_axis, dependent_axis in (
                (Axes.X, Coordinates.X),
                (Axes.Y, Coordinates.Y),
                (Axes.ZPLANE, Coordinates.Z),
        ):
            assert len(sel_xarray[primary_axis.value]) == len(sel_xarray[dependent_axis.value])


def test_select_and_export():
    """Tests selecting on an Imagestack with a shape (5, 5, 15, 200, 200)
        1.) stack.sel(indexers)
        2.) export stack
    """
    stack = synthetic_stack(
        num_round=5, num_ch=5, num_z=15, tile_height=200, tile_width=200)

    # select on range of rounds and single ch and Z
    selected = stack.sel({Axes.ROUND: (1, None), Axes.CH: (2, 3), Axes.ZPLANE: 0})

    with tempfile.TemporaryDirectory() as tfd:
        path = os.path.join(tfd, "stack.json")
        selected.export(path)

        loaded = ImageStack.from_path_or_url(path)

        assert np.array_equal(selected.xarray, loaded.xarray)
        for coords in (
                Axes.ROUND, Axes.CH, Axes.ZPLANE, Coordinates.X, Coordinates.Y, Coordinates.Z):
            assert np.allclose(selected.xarray[coords.value], loaded.xarray[coords.value])



import warnings
from itertools import product
from typing import Tuple

import numpy as np
from scipy.ndimage.filters import gaussian_filter
from skimage import img_as_float32, img_as_uint

from starfish import Codebook, ImageStack, IntensityTable
from starfish.core.image.Filter.white_tophat import WhiteTophat
from starfish.core.imagestack.test.factories import create_imagestack_from_codebook
from starfish.core.spots.DecodeSpots import MetricDistance
from starfish.core.spots.FindSpots import BlobDetector
from starfish.core.types import Axes, Features


class SyntheticData:
    """Synthetic generator for spot data and container objects

    Currently, this generator only generates codebooks with single channels "on" in each
    imaging round. Therefore, it can generate ISS and smFISH type experiments, but not
    more complex codes.

    Examples
    --------
    >>> from starfish.core.test.factories import SyntheticData

    >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2)
    >>> codes = sd.codebook()
    >>> codes
    <xarray.Codebook (target: 2, c: 3, h: 4)>
    array([[[0, 0, 0, 0],
            [0, 0, 1, 1],
            [1, 1, 0, 0]],

           [[1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 1]]], dtype=uint8)
    Coordinates:
      * target     (target) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...
      * c          (c) int64 0 1 2
      * h          (h) int64 0 1 2 3

    >>> # note that the features are drawn from the codebook
    >>> intensities = sd.intensities(codebook=codes)
    >>> intensities
    <xarray.IntensityTable (features: 2, c: 3, h: 4)>
    array([[[    0.,     0.,     0.,     0.],
            [    0.,     0.,  8022., 12412.],
            [11160.,  9546.,     0.,     0.]],

           [[    0.,     0.,     0.,     0.],
            [    0.,     0., 10506., 10830.],
            [11172., 12331.,     0.,     0.]]])
    Coordinates:
      * features   (features) MultiIndex
      - z          (features) int64 7 3
      - y          (features) int64 14 32
      - x          (features) int64 32 15
      - r          (features) float64 nan nan
      * c          (c) int64 0 1 2
      * h          (h) int64 0 1 2 3
        target     (features) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...

    >>> sd.spots(intensities=intensities)
    <starfish.imagestack.imagestack.ImageStack at 0x10a60c5f8>

    """

    def __init__(
            self,
            n_round: int = 4,
            n_ch: int = 4,
            n_z: int = 10,
            height: int = 50,
            width: int = 45,
            n_codes: int = 16,
            n_spots: int = 20,
            n_photons_background: int = 1000,
            background_electrons: int = 1,
            point_spread_function: Tuple[int, ...] = (4, 2, 2),
            camera_detection_efficiency: float = 0.25,
            gray_level: float = 37000.0 / 2 ** 16,
            ad_conversion_bits: int = 16,
            mean_fluor_per_spot: int = 200,
            mean_photons_per_fluor: int = 50,

    ) -> None:
        self.n_round = n_round
        self.n_ch = n_ch
        self.n_z = n_z
        self.height = height
        self.width = width
        self.n_codes = n_codes
        self.n_spots = n_spots
        self.n_photons_background = n_photons_background
        self.background_electrons = background_electrons
        self.point_spread_function = point_spread_function
        self.camera_detection_efficiency = camera_detection_efficiency
        self.gray_level = gray_level
        self.ad_coversion_bits = ad_conversion_bits
        self.mean_fluor_per_spot = mean_fluor_per_spot
        self.mean_photons_per_fluor = mean_photons_per_fluor

    def codebook(self) -> Codebook:
        return Codebook.synthetic_one_hot_codebook(self.n_round, self.n_ch, self.n_codes)

    def intensities(self, codebook=None) -> IntensityTable:
        if codebook is None:
            codebook = self.codebook()
        intensities = IntensityTable.synthetic_intensities(
            codebook, self.n_z, self.height, self.width, self.n_spots,
            self.mean_fluor_per_spot, self.mean_photons_per_fluor)
        assert intensities.dtype == np.float32 and intensities.max() <= 1
        return intensities

    def spots(self, intensities=None) -> ImageStack:
        if intensities is None:
            intensities = self.intensities()
        stack = SyntheticData.synthetic_spots(
            intensities, self.n_z, self.height, self.width, self.n_photons_background,
            self.point_spread_function, self.camera_detection_efficiency,
            self.background_electrons, self.gray_level, self.ad_coversion_bits)
        assert stack.xarray.dtype == np.float32
        return stack

    @classmethod
    def synthetic_spots(
            cls, intensities: IntensityTable, num_z: int, height: int, width: int,
            n_photons_background=1000, point_spread_function=(4, 2, 2),
            camera_detection_efficiency=0.25, background_electrons=1,
            graylevel: float = 37000.0 / 2 ** 16, ad_conversion_bits=16,
    ) -> ImageStack:
        """Generate a synthetic ImageStack from a set of Features stored in an IntensityTable

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable containing coordinates of fluorophores. Used to position and generate
            spots in the output ImageStack
        num_z : int
            Number of z-planes in the ImageStack
        height : int
            Height in pixels of the ImageStack
        width : int
            Width in pixels of the ImageStack
        n_photons_background : int
            Poisson rate for the number of background photons to add to each pixel of the image.
            Set this parameter to 0 to eliminate background.
            (default 1000)
        point_spread_function : Tuple[int]
            The width of the gaussian density wherein photons spread around their light source.
            Set to zero to eliminate this (default (4, 2, 2))
        camera_detection_efficiency : float
            The efficiency of the camera to detect light. Set to 1 to remove this filter (default
            0.25)
        background_electrons : int
            Poisson rate for the number of spurious electrons detected per pixel during image
            capture by the camera (default 1)
        graylevel : float
            The number of shades of gray displayable by the synthetic camera. Larger numbers will
            produce higher resolution images (default 37000 / 2 ** 16)
        ad_conversion_bits : int
            The number of bits used during analog to visual conversion (default 16)

        Returns
        -------
        ImageStack :
            synthetic spots

        """
        # check some params
        if not 0 < camera_detection_efficiency <= 1:
            raise ValueError(
                f'invalid camera_detection_efficiency value: {camera_detection_efficiency}. '
                f'Must be in the interval (0, 1].')

        def select_uint_dtype(array):
            """choose appropriate dtype based on values of an array"""
            max_val = np.max(array)
            for dtype in (np.uint8, np.uint16, np.uint32):
                if max_val <= np.iinfo(dtype).max:
                    return array.astype(dtype)
            raise ValueError('value exceeds dynamic range of largest skimage-supported type')

        # make sure requested dimensions are large enough to support intensity values
        axis_to_size = zip((Axes.ZPLANE.value, Axes.Y.value, Axes.X.value), (num_z, height, width))
        for axis, requested_size in axis_to_size:
            required_size = intensities.coords[axis].values.max() + 1
            if required_size > requested_size:
                raise ValueError(
                    f'locations of intensities contained in table exceed the size of requested '
                    f'axis {axis}. Required size {required_size} > {requested_size}.')

        # create an empty array of the correct size
        image = np.zeros(
            (
                intensities.sizes[Axes.ROUND.value],
                intensities.sizes[Axes.CH.value],
                num_z,
                height,
                width
            ), dtype=np.uint32
        )

        # starfish uses float images, but the logic here requires uint. We cast, and will cast back
        # at the end of the function
        intensities.values = img_as_uint(intensities)

        for round_, ch in product(*(range(s) for s in intensities.shape[1:])):
            spots = intensities[{Axes.ROUND.value: round_, Axes.CH.value: ch}]

            # numpy deprecated casting a specific way of casting floats that is triggered in xarray
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FutureWarning)
                values = spots.where(spots, drop=True)

            image[round_, ch, values.z, values.y.astype(int), values.x.astype(int)] = values

        intensities.values = img_as_float32(intensities)

        # add imaging noise
        image += np.random.poisson(n_photons_background, size=image.shape).astype(np.uint32)

        # blur image over coordinates, but not over round_/channels (dim 0, 1)
        sigma = (0, 0) + point_spread_function
        image = gaussian_filter(image, sigma=sigma, mode='nearest')

        image = image * camera_detection_efficiency

        image += np.random.normal(scale=background_electrons, size=image.shape)

        # mimic analog to digital conversion
        image = (image / graylevel).astype(int).clip(0, 2 ** ad_conversion_bits)

        # clip in case we've picked up some negative values
        image = np.clip(image, 0, a_max=None)

        # set the smallest int datatype that supports the data's intensity range
        image = select_uint_dtype(image)

        # convert to float for ImageStack
        with warnings.catch_warnings():
            # possible precision loss when casting from uint to float is acceptable
            warnings.simplefilter('ignore', UserWarning)
            image = img_as_float32(image)

        return ImageStack.from_numpy(image)


def two_spot_one_hot_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 2-channel 2-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    The encoding of this data is similar to that used in In-situ Sequencing, FISSEQ,
    BaristaSeq, STARMAP, MExFISH, or SeqFISH.

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    float :
        the maximum intensity found in the created ImageStack

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity


def two_spot_sparse_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 3-channel 3-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    These spots display sparsity in both rounds and channels, similar to the sparse encoding of
    MERFISH

    Returns
    -------
    ImageStack :
        noiseless ImageStack containing two spots

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 2, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity


def two_spot_informative_blank_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 4-channel 2-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    The encoding of this data is essentially a one-hot encoding, but where one of the channels is a
    intentionally and meaningfully "blank".

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    float :
        the maximum intensity found in the created ImageStack

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                # round 3 is blank and channel 3 is not used
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                # round 0 is blank and channel 0 is not used
                {Axes.ROUND.value: 1, Axes.CH.value: 3, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 2, Features.CODE_VALUE: 1},
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = create_imagestack_from_codebook(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity


def synthetic_dataset_with_truth_values():
    np.random.seed(2)
    synthesizer = SyntheticData(n_spots=5)
    codebook = synthesizer.codebook()
    true_intensities = synthesizer.intensities(codebook=codebook)
    image = synthesizer.spots(intensities=true_intensities)

    return codebook, true_intensities, image


def synthetic_dataset_with_truth_values_and_called_spots():
    codebook, true_intensities, image = synthetic_dataset_with_truth_values()

    wth = WhiteTophat(masking_radius=15)
    filtered = wth.run(image, in_place=False)

    min_sigma = 1.5
    max_sigma = 4
    num_sigma = 10
    threshold = 1e-4
    gsd = BlobDetector(min_sigma=min_sigma,
                       max_sigma=max_sigma,
                       num_sigma=num_sigma,
                       threshold=threshold,
                       measurement_type='max')
    spots = gsd.run(image_stack=filtered)
    decoder = MetricDistance(codebook=codebook, max_distance=1, min_intensity=0, norm_order=2)
    decoded_intensities = decoder.run(spots)
    assert decoded_intensities.shape[0] == 5
    return codebook, true_intensities, image, decoded_intensities


def synthetic_spot_pass_through_stack():
    codebook, true_intensities, _ = synthetic_dataset_with_truth_values()
    true_intensities = true_intensities[:2]
    # transfer the intensities to the stack but don't do anything to them.
    img_stack = SyntheticData.synthetic_spots(
        true_intensities, num_z=12, height=50, width=45, n_photons_background=0,
        point_spread_function=(0, 0, 0), camera_detection_efficiency=1.0,
        background_electrons=0, graylevel=1)
    return codebook, true_intensities, img_stack


def codebook_intensities_image_for_single_synthetic_spot():
    sd = SyntheticData(
        n_round=2, n_ch=2, n_z=2, height=20, width=30, n_codes=1, n_spots=1
    )
    codebook = sd.codebook()
    intensities = sd.intensities(codebook)
    image = sd.spots(intensities)
    return codebook, intensities, image



#!/usr/bin/env python
import subprocess
import sys

import pkg_resources

from starfish.core.spacetx_format.cli import validate as validate_cli
from starfish.core.util import click


def art_string():
    return r"""
         _              __ _     _
        | |            / _(_)   | |
     ___| |_ __ _ _ __| |_ _ ___| |__
    / __| __/ _` | '__|  _| / __| '_  `
    \__ \ || (_| | |  | | | \__ \ | | |
    |___/\__\__,_|_|  |_| |_|___/_| |_|

    """


@click.group()
@click.pass_context
def starfish(ctx):
    """
    standardized analysis pipeline for image-based transcriptomics
    see: https://spacetx-starfish.readthedocs.io for more information.
    """
    art = art_string()
    print_art = True
    sub = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    if hasattr(sub, "no_art"):
        print_art = not sub.no_art
    if print_art:
        print(art)


@starfish.command()
def version():
    from starfish import __version__
    print(__version__)
version.no_art = True  # type: ignore


@starfish.group()
def util():
    """
    house-keeping commands for the starfish library
    """
    pass


@util.command()
def install_strict_dependencies():
    """
    warning! updates different packages in your local installation
    """
    strict_requirements_file = pkg_resources.resource_filename(
        "starfish", "REQUIREMENTS-STRICT.txt")
    subprocess.check_call([
        sys.executable, "-m", "pip", "install", "-r", strict_requirements_file
    ])


# Other
starfish.add_command(validate_cli)  # type: ignore



from typing import TypeVar

import numpy as np
import xarray as xr


PreserveFloatRangeType = TypeVar("PreserveFloatRangeType", xr.DataArray, np.ndarray)
LevelsType = TypeVar("LevelsType", xr.DataArray, np.ndarray)


def preserve_float_range(
        array: PreserveFloatRangeType,
        rescale: bool = False,
        preserve_input: bool = False,
) -> PreserveFloatRangeType:
    """
    Clips values below zero to zero. If values above one are detected, clips them
    to 1 unless `rescale` is True, in which case the input is scaled by
    the max value and the linearity of the dynamic range is preserved.

    Input arrays may be modified by this operation if `preserve_input` is False.  There is no
    guarantee that the input array is returned even if `preserve_input` is True, however.

    Parameters
    ----------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values should be in the interval [0, 1] but may not be.
    rescale : bool
        If true, scale values by the max.
    preserve_input : bool
        If True, ensure that we do not modify the input data.  This may either be done by making a
        copy of the input data or ensuring that the operation does not modify the input array.

        Even if `preserve_input` is True, modifications to the resulting array may modify the input
        array.

    Returns
    -------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values are in the interval [0, 1].

    """
    return levels(array, rescale=False, rescale_saturated=rescale, preserve_input=preserve_input)


def levels(
        array: LevelsType,
        rescale: bool = False,
        rescale_saturated: bool = False,
        preserve_input: bool = False,
) -> LevelsType:
    """
    Clips values below zero to zero.  If values above one are detected, clip them to 1 if both
    ``rescale`` and ``rescale_saturated`` are False.  If ``rescale`` is True, then the input is
    rescaled by the peak intensity.  If ``rescale_saturated`` is True, then the image is rescaled
    by the peak intensity provided the peak intensity is saturated.  It is illegal for both
    ``rescale`` and ``rescale_saturated`` to be True.

    Input arrays may be modified by this operation if `preserve_input` is False.  There is no
    guarantee that the input array is returned even if `preserve_input` is True, however.

    Parameters
    ----------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values should be in the interval [0, 1] but may not be.
    rescale : bool
        If true, scale values by the max.
    rescale_saturated : bool
        If true, scale values by the max if the max is saturated (i.e., >= 1).
    preserve_input : bool
        If True, ensure that we do not modify the input data.  This may either be done by making a
        copy of the input data or ensuring that the operation does not modify the input array.

        Even if `preserve_input` is True, modifications to the resulting array may modify the input
        array.

    Returns
    -------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values are in the interval [0, 1].
    """
    if rescale and rescale_saturated:
        raise ValueError("rescale and rescale_saturated cannot both be set.")

    if isinstance(array, xr.DataArray):
        # do a shallow copy
        array = array.copy(deep=False)
        data = array.values
    else:
        data = array

    casted = data.astype(np.float32, copy=False)
    if casted is not data:
        preserve_input = False
    data = casted

    if preserve_input or not data.flags['WRITEABLE']:
        # if we still want a copy, check to see if any modifications would be made.  if so, make a
        # copy.
        belowzero = np.any(data < 0)
        aboveone = np.any(data > 1)
        # we could potentially skip the copy if it's just 'rescale' and the max is exactly 1.0, but
        # that is a pretty unlikely scenario, so we assume the max is not going to be exactly 1.0.
        if belowzero or aboveone or rescale:
            data = data.copy()
            _adjust_image_levels_in_place(data, rescale or (rescale_saturated and aboveone))
    else:
        # we don't want a copy, so we just do it in place.
        aboveone = np.any(data > 1)
        _adjust_image_levels_in_place(data, rescale or (rescale_saturated and aboveone))

    if isinstance(array, xr.DataArray):
        if data is not array.values:
            array.values = data
    else:
        array = data

    return array


def _adjust_image_levels_in_place(
        array: np.ndarray,
        rescale: bool = False,
):
    """
    Clips values below zero to zero.  If rescale is True, the input is scaled by the max value and
    the linearity of the dynamic range is preserved.

    Parameters
    ----------
    array : np.ndarray
        Array whose values should be in the interval [0, 1] but may not be.  This array must be of
        type np.float32.
    rescale : bool
        If true, scale values by the max.

    Returns
    -------
    array : np.ndarray
        Array whose values are in the interval [0, 1].

    """
    assert array.dtype == np.float32

    array[array < 0] = 0
    if rescale:
        array /= array.max()
    else:
        array[array > 1] = 1
